import gym
from fastai.basic_train import LearnerCallback, DatasetType
from fastai.tabular.data import def_emb_sz
from gym import Wrapper
from gym.spaces import Discrete, Box, MultiDiscrete, Dict

from fast_rl.core.basic_train import AgentLearner
from fast_rl.util.exceptions import MaxEpisodeStepsMissingError
import os

# Some imported libraries have env wrappers that can make compatibility less messy.
WRAP_ENV_FNS = []
# Because concurrency errors happen from Open AI when there are multiple environments.
os.environ['KMP_DUPLICATE_LIB_OK'] = 'True'

try:
    import pybullet
    import pybulletgym.envs
    from pybulletgym.envs.mujoco.envs.env_bases import BaseBulletEnv as MujocoEnv
    from pybulletgym.envs.roboschool.envs.env_bases import BaseBulletEnv as RoboschoolEnv

    class BulletWrapper(Wrapper):
        def step(self, action):
            r""" In the event of a random disconnect, retry the env """
            try:
                result = super().step(action)
            except pybullet.error as e:
                self.__init__(env=gym.make(self.spec.id))
                super().reset()
                result = super().step(action)
            return result

    def pybullet_wrap(env, render):
        if issubclass(env.unwrapped.__class__, (MujocoEnv, RoboschoolEnv)):
            env = BulletWrapper(env=env)
            if render == 'human': env.render()
        return env

    WRAP_ENV_FNS.append(pybullet_wrap)

except ModuleNotFoundError as e:
    print(f'Can\'t import one of these: {e}')
try:
    # noinspection PyUnresolvedReferences
    import gym_maze
    from gym_maze.envs.maze_env import MazeEnv
    import pygame
    # import importlib

    class GymMazeWrapper(Wrapper):
        #     pygame.init()
        #
        # def render(self, mode='human', **kwargs):
        #     try:
        #         return self.env.render(mode=mode, **kwargs)
        #     except pygame.error as e:
        #         original_id = self.env.spec.id
        #         del self.env
        #         pygame.init()
        #         self.env = gym.make(original_id)
        #         super().reset()
        #         return self.env.render(mode=mode, **kwargs)

        def close(self):
            self.env.maze_view.quit_game()
            self.env.unwrapped.enable_render = False


    def gymmaze_wrap(env, render):
        if issubclass(env.unwrapped.__class__, MazeEnv):
            env = GymMazeWrapper(env=env)
        return env

    WRAP_ENV_FNS.append(gymmaze_wrap)

except ModuleNotFoundError as e:
    print(f'Can\'t import one of these: {e}')
try:
    # noinspection PyUnresolvedReferences
    import gym_minigrid
    # noinspection PyUnresolvedReferences
    from gym_minigrid.minigrid import MiniGridEnv
    # noinspection PyUnresolvedReferences
    from gym_minigrid.wrappers import FlatObsWrapper

    def mini_grid_wrap(env, **kwargs):
        if issubclass(env.__class__, MiniGridEnv): env = FlatObsWrapper(env)
        return env

    WRAP_ENV_FNS.append(mini_grid_wrap)
except ModuleNotFoundError as e:
    print(f'Can\'t import one of these: {e}')

from fastai.core import *
from fastai.data_block import ItemList, Tensor, Dataset, DataBunch, DataLoader
from fastai.imports import torch
from datetime import datetime
import pickle

from fast_rl.util.misc import b_colors, list_in_str

FEED_TYPE_IMAGE = 0
FEED_TYPE_STATE = 1


@dataclass
class Bounds(object):
    r"""
    Handles bounds for 1 dimensional spaces.

    between (gym.Space): A convenience variable for determining the min and max bounds

    discrete (bool): Whether the Bounds are discrete or not. Important for N possible v calc.

    min (list): Correlated min for a given dimension

    max (list): Correlated max for a given dimension
    """
    between: Any = None
    discrete: bool = False
    min: list = None
    max: list = None

    def __len__(self):
        r"""
        Returns the number of dimensions the bounds have.

        `self.min`'s length is returned because `self.max` and `self.min` as validated to have the same length.
        """
        return len(self.min)

    @property
    def n_possible_values(self):
        """
        Returns the maximum number of v that can be taken.

        This is important for doing embeddings.
        """
        if not self.discrete: return np.inf
        else: return np.add(*np.abs((self.max, self.min))).reshape(1, -1)

    def __post_init__(self):
        """Sets min and max fields then validates them."""
        self.min, self.max = ifnone(self.min, []), ifnone(self.max, [])
        # If a tuple has been passed, break it into correlated min max variables.
        if self.between is not None:
            between = list(self.between.spaces.values()) if isinstance(self.between, gym.spaces.Dict) else self.between

            for b in (between if isinstance(between, gym.spaces.Tuple) else listify(between)):
                if isinstance(b, (int, np.int64, np.int, float, np.float)):
                    self.min, self.max = self.min + [0], self.max + [b]
                    self.discrete = isinstance(b, (int, np.int, np.int64))
                elif isinstance(b, Discrete):
                    self.min, self.max, self.discrete = self.min + [0], self.max + [b.n], True
                elif isinstance(b, MultiDiscrete):
                    self.min, self.max, self.discrete = self.min + [0] * sum(b.nvec.shape), self.max + b.nvec, True
                elif isinstance(b, Box):
                    self.min, self.max = self.min + list(b.low), self.max + list(b.high)
                    self.discrete = any([b.dtype in (int, np.int, np.int64)])
                else:
                    raise ValueError(f'Tuple not understood {self.between}')

        if len(self.min) != len(self.max):
            raise ValueError(f'Min Max do not match min {len(self.min)} max {len(self.max)}')
        if len(self.min) == 0: raise ValueError(f'Min and Max are 0')


@dataclass
class Action(object):
    r"""
    Handles actions, action space, and value verification.

    An important difference between taken_action and raw_action is that the raw action is the immediate
    raw output from the model before any argmax processing.

    taken_action (np.array): Expected to always to be numpy arrays with shape (-1, 1). This is the action that
    is to be input into the env `step` function.

    raw_action (np.array): Expected to always to be numpy arrays with shape (-1, 1). This is the raw model
    output such as neural net final layer output. Can be None.

    action_space (gym.Space): Used for estimating the max number of v. This is important for embeddings.

    bounds (tuple): Maximum and minimum v for each action dimension.

    n_possible_values (int): An integer or inf value indicating the total number of possible actions there are.
    """
    taken_action: torch.tensor
    action_space: gym.Space
    raw_action: torch.tensor = None
    bounds: Bounds = None

    def to(self, device):
        self.taken_action = self.taken_action.to(device=device)
        if self.raw_action is not None: self.raw_action = self.raw_action.to(device=device)

    @property
    def n_possible_values(self): return self.bounds.n_possible_values

    def __post_init__(self):
        # Determine bounds
        self.bounds = Bounds(self.action_space)

        # Fix shapes
        if not isinstance(self.taken_action, torch.Tensor):
            self.taken_action = torch.tensor(data=self.taken_action).view(1, -1)
        else: self.taken_action = self.taken_action.view(1, -1)
        if self.raw_action is not None: self.raw_action = torch.tensor(data=self.raw_action).reshape(1, -1)

    def get_single_action(self):
        """ OpenAI envs do not like 1x1 arrays when they are expecting scalars, so we need to unwrap them. """
        a = self.taken_action.detach().numpy()[0]
        if len(self.bounds) == 1: a = a[0]
        if self.bounds.discrete and len(self.bounds) == 1: return int(a)
        elif self.bounds.discrete and len(self.bounds) != 1: return a.astype(int)
        elif not self.bounds.discrete and len(self.bounds) == 1: return [float(a)]
        elif not self.bounds.discrete and len(self.bounds) != 1: return a.reshape(-1,).astype(float)
        raise ValueError(f'This should not have crashed.')

    def set_single_action(self, action: np.array):
        if np.isscalar(action):
            self.taken_action = torch.tensor(data=action).reshape(1, -1)
        elif len(action.shape) == 1:
            self.taken_action = torch.tensor(data=action).reshape(1, -1)
        elif len(action.shape) == 3 and action.shape[0] == 1:
            self.taken_action = torch.tensor(data=action)[0]
        self.taken_action = self.taken_action.int() if self.bounds.discrete else self.taken_action.float()


@dataclass
class State(object):
    r"""
    Handles states, both their main and alternate formats.

    s (np.array): State space acquired from `env.reset` `s_prime` or `env.render`.

    s_prime (np.array): State space acquired from `env.step` or `env.render`.

    alt_s (np.array): Alternate State space acquired from `env.reset` `s_prime` or `env.render`. Should be an image.

    alt_s_prime (np.array): Alternate State space acquired from `env.step` or `env.render`. Should be an image.

    mode (int): Should be either FEED_TYPE_IMAGE or FEED_TYPE_STATE

    observation_space (gym.Space): Used for estimating the max number of v. This is important for embeddings.

    bounds (Bounds): Maximum and minimum v for each state dimension.

    n_possible_values (int): An integer or inf value indicating the total number of possible actions there are.
    """
    s: torch.tensor
    s_prime: torch.tensor
    alt_s: Union[torch.tensor, np.array]
    alt_s_prime: Union[torch.tensor, np.array]
    observation_space: gym.Space
    mode: int = FEED_TYPE_STATE
    bounds: Bounds = None

    def to(self, device):
        self.s = self.s.to(device=device)
        self.s_prime = self.s_prime.to(device=device)

    @property
    def channels(self):
        return self.s.shape[3]

    @property
    def n_possible_values(self): return self.bounds.n_possible_values

    def __str__(self):
        out = copy(self.__dict__)
        for key in out:
            if out[key] is not None and (key == 's' or list_in_str(key, ['_s', 's_'])):  out[key] = out[key].shape

        return f'State: ' + ', '.join([str(i) for i in out.items()])

    def _fix_field(self, input_field):
        if input_field is None: return None
        input_field = copy(input_field)

        if type(input_field) is str: input_field = np.array(input_field)
        elif type(input_field) is tuple:
            dtype = int if self.bounds.discrete else float
            input_field = torch.tensor(data=np.array(input_field).reshape(1, -1).astype(dtype))
        elif np.isscalar(input_field): input_field = torch.tensor(data=input_field)
        elif type(input_field) is torch.Tensor: input_field = input_field.clone().detach()
        else: input_field = torch.from_numpy(input_field)

        input_field = input_field.long() if self.bounds.discrete and self.mode != FEED_TYPE_IMAGE else input_field.float()
        # If a non-image state missing the batch dim
        if len(input_field.shape) <= 1: return input_field.reshape(1, -1)
        # If a non-image 2+d state missing the batch dim
        elif input_field.shape[0] != 1 and len(input_field.shape) != 3: return input_field.reshape(1, -1)
        # If an image state and missing the batch dim
        elif input_field.shape[0] != 1 and len(input_field.shape) == 3: return input_field.unsqueeze(0)
        # If an image with 4 dims (b, w, h, c), return safely
        elif input_field.shape[0] == 1 and len(input_field.shape) == 4: return input_field
        # If not an image, but has 2 dims (b, d), return safely
        elif input_field.shape[0] == 1 and len(input_field.shape) == 2: return input_field
        raise ValueError(f'Input has shape {input_field} for mode {self.mode}. This is unexpected.')

    def __post_init__(self):
        if self.mode not in [FEED_TYPE_IMAGE, FEED_TYPE_STATE]:
            ValueError(f'Mode invalid {self.mode} not valid feed type')
        # We want to swap the state variables if the alt is the image state
        if self.mode == FEED_TYPE_IMAGE and len(self.alt_s.shape) > 2 and len(self.s.shape) != 3:
            self.observation_space = gym.spaces.Box(0, 255, self.alt_s.shape, dtype=np.int64)
            self.alt_s, self.alt_s_prime, self.s, self.s_prime = self.s, self.s_prime, self.alt_s, self.alt_s_prime
        # Determine bounds
        self.bounds = Bounds(self.observation_space)
        # Fix Shapes
        self.s, self.s_prime = self._fix_field(self.s), self._fix_field(self.s_prime)
        self.alt_s, self.alt_s_prime = self._fix_field(self.alt_s), self._fix_field(self.alt_s_prime)


@dataclass
class MDPStep(object):
    r"""
    Contains all the variables to represent a Markov Decision Process step.

    action (Action):

    state (State):

    done (bool):

    reward (float):

    episode (int):

    step (int):

    """
    action: Action
    state: State
    done: torch.tensor
    reward: torch.tensor
    episode: int
    step: int

    def __post_init__(self):
        self.action = deepcopy(self.action)
        self.state = deepcopy(self.state)
        self.reward = torch.tensor(data=self.reward).reshape(1, -1).float()
        self.done = torch.tensor(data=self.done).reshape(1, -1).float()

    def to(self, device):
        self.reward = self.reward.to(device=device)
        self.done = self.done.to(device=device)
        self.action.to(device=device)
        self.state.to(device=device)

    def __str__(self):
        return ', '.join([str(self.__dict__[el]) for el in self.__dict__])

    def clean(self):
        r""" Removes fields that are generally unimportant (purely debugging) """
        self.state.alt_s_prime = None
        self.state.alt_s = None
        self.state.observation_space = None
        self.state.bounds = None
        self.action.raw_action = None
        self.action.action_space = None
        self.action.bounds = None

    @property
    def data(self): return self.s_prime[0], self.alt_s_prime[0] if self.alt_s_prime is not None else None
    @property
    def obj(self): return self.__dict__
    @property
    def s(self): return self.state.s
    @property
    def s_prime(self): return self.state.s_prime
    @property
    def alt_s_prime(self): return self.state.alt_s_prime
    @property
    def a(self): return self.action.taken_action
    @property
    def d(self):
        return bool(self.done)


class MDPCallback(LearnerCallback):
    _order = -11  # Needs to happen before Recorder

    @property
    def learn(self) -> AgentLearner: return self._learn()

    def __init__(self, learn):
        r"""
        Handles action assignment, episode naming.

        Args:
            learn:
        """
        super().__init__(learn)
        self.train_ds: MDPDataset = learn.data.train_ds
        self.valid_ds: MDPDataset = None if learn.data.empty_val else learn.data.valid_ds

    def on_batch_begin(self, last_input, last_target, train, **kwargs: Any):
        r""" Set the Action of a dataset, determine if still warming up. """
        a = self.learn.predict(last_input)
        if self.learn.model.training:
            self.train_ds.action = Action(taken_action=a, action_space=self.train_ds.action.action_space)
        else: self.valid_ds.action = Action(taken_action=a, action_space=self.train_ds.action.action_space)
        self.train_ds.is_warming_up = self.learn.warming_up
        if self.valid_ds is not None: self.valid_ds.is_warming_up = self.learn.warming_up
        if not self.learn.warming_up and self.learn.loss_func is None:
            self.learn.init_loss_func()
        return {'skip_bwd': True, 'train': not self.train_ds.is_warming_up and train}

    def on_backward_end(self, **kwargs: Any):
        return {'skip_step': True}

    def on_step_end(self, **kwargs: Any):
        return {'skip_zero': True}

    def on_epoch_end(self, last_metrics, epoch, **kwargs: Any) -> None:
        """ Updates the most recent episode number in both datasets. """
        self.train_ds.x.set_recent_run_episode(epoch)
        self.train_ds.episode = epoch
        if last_metrics[0] is not None and self.valid_ds is not None:
            self.valid_ds.x.set_recent_run_episode(epoch)
            self.valid_ds.episode = epoch

    def on_train_end(self, **kwargs:Any) ->None:
        self.learn.data.close()


class MDPMemoryManager(LearnerCallback):
    def __init__(self, learn, strategy, k=1):
        super().__init__(learn)
        self.strategy = strategy
        self.k = k
        self._strategy_fn_dict = {
            'k_partitions_top': self.k_top
        }

    def _comp_less(self, key, info, episode):
        return info[key] < info[episode]

    def _comp_greater(self, key, info, episode):
        return info[key] > info[episode]

    def k_top(self, info: Dict[int, List[Tuple[float, bool]]]):
        # If the episode -1 is defined, then clean it, this is just placeholder data
        if -1 in info and not info[-1][1]: return [-1]
        # If the number of not-clean episodes is less than k, don't do anything
        if len([k for k in info if not info[k][1]]) <= self.k: return None
        # Collect k top episodes, then clean the rest
        k_top = []
        for episode in [i for i in info if i != -1]:
            # If the episode is greater than all of the other episodes that are not already in k_top
            compared = [info[episode][0] > info[k][0] for k in info if k not in k_top and k != -1 and k != episode]
            if all(compared) and len(compared) != 0 and len(k_top) < self.k:
                k_top.append(episode)

        return list(set([k for k in info if not info[k][1]]) - set(k_top))

    def on_epoch_end(self, **kwargs: Any):
        for ds_type in [DatasetType.Train] if self.learn.data.empty_val else [DatasetType.Train, DatasetType.Valid]:
            ds: MDPDataset = self.learn.dl(ds_type).dataset
            episodes = self._strategy_fn_dict[self.strategy](ds.x.info)
            if episodes is not None:
                for e in episodes: ds.x.clean(e)


class MDPDataset(Dataset):
    def __init__(self, env: gym.Env, memory_manager, bs, render='rgb_array', feed_type=FEED_TYPE_STATE, max_steps=None,
                 x=None):
        r"""
        Handles env execution and ItemList building.

        Args:
            env: OpenAI environment to execute.
            memory_manager: Handles how the list size will be reduced sch as removing image data.
            bs: Size of a single batch for models and the dataset to use.
        """
        for wrapper_fn in WRAP_ENV_FNS: env = wrapper_fn(env, render=render)
        self.env = env
        self.render = render
        self.feed_type = feed_type
        self.bs = bs
        self._max_steps = max_steps
        self.action = Action(taken_action=self.env.action_space.sample(), action_space=self.env.action_space)
        self.state = None
        self.s_prime, self.alt_s_prime = None, None
        self.callback = [MDPCallback, memory_manager]
        # Tracking fields
        self.episode = -1 if x is None else max([i.episode + 1 for i in x.items])
        self.counter = 0
        # While true, the dataset object with loop until the the number of loops is more than the batch size.
        # This allows a model to fill its buffers before doing proper epoch iterating.
        self.is_warming_up = True

        # FastAI fields
        self.x = ifnone(x, MDPList([]))
        self.item: Union[MDPStep, None] = None
        self.new(None)

    def get_emb_szs(self):
        return [def_emb_sz(0, 0,None)]

    def aug_steps(self, steps):
        if self.is_warming_up and steps < self.bs: return self.bs
        return steps

    @property
    def max_steps(self):
        if self._max_steps is not None: return self._max_steps
        if hasattr(self.env, 'max_steps'): return getattr(self.env, 'max_steps')
        if hasattr(self.env.unwrapped, 'max_steps'): return getattr(self.env.unwrapped, 'max_steps')
        if hasattr(self.env, '_max_episode_steps'): return getattr(self.env, '_max_episode_steps')
        if self.env.spec.max_episode_steps is not None: return self.env.spec.max_episode_steps

        msg = f'Env {self.env.spec.id} does not have max episode steps. '
        msg += ' Either pass the max steps as a param, or catch this exception then pass a default max step param.'
        raise MaxEpisodeStepsMissingError(msg)

    @property
    def image(self):
        r""" Needed because of blackjack-v0 env >:( """
        try:
            current_image = self.env.render('rgb_array')
            if self.render == 'human': self.env.render(self.render)
        except NotImplementedError:
            print(f'{b_colors.WARNING} {self.env.unwrapped.spec} Not returning Image {b_colors.ENDC}')
            current_image = None
        return current_image

    def __del__(self):
        self.env.close()

    def __len__(self):
        return self.aug_steps(self.max_steps)

    def stage_1_env_reset(self):
        r"""
        Handles environment resetting and dataset batch termination.

        We are interested in the entire dataset ending when an item is done.

        Returns: The state space and the image after a reset.

        """
        if self.counter != 0 and self.item.d:
            self.counter = 0
            if not self.is_warming_up:
                self.env.reset()
                raise StopIteration
        if self.item is None or self.item.d:
            return self.env.reset(), self.image
        return self.s_prime, self.alt_s_prime

    def stage_2_env_step(self) -> Tuple[np.array, float, bool, None, np.array]:
        r"""
        Handles taking a step in the environment.

        We want to cancel the env early if we are at our max step amount.

        Returns: The state, reward, whether the episode is done, and the image.
        """
        s_prime, reward, done, _ = self.env.step(self.action.get_single_action())
        # If we are at the max steps limit but the env is not done, we need to force an env end.
        # However, we need the loop to iterate +1 time for allowing the stage_1_env_reset
        if len(self) - 2 == self.counter: done = True
        return s_prime, reward, done, _, self.image

    def new(self, _):
        s, alt_s = self.stage_1_env_reset()
        self.s_prime, reward, done, _, self.alt_s_prime = self.stage_2_env_step()
        # If both the current item and the done are both true, then we need to retry the env
        if self.item is not None and self.item.d and done: return self.new()

        self.state = State(s, self.s_prime, alt_s, self.alt_s_prime,  self.env.observation_space, self.feed_type)
        self.item = MDPStep(self.action, self.state, done, reward, self.episode, self.counter)
        self.counter += 1

        return MDPList([self.item])

    def __getitem__(self, _):
        item = self.new(_)
        self.x.add(item)
        return self.x[-1]

    def to_csv(self, root_path, name):
        if not os.path.exists(root_path): os.makedirs(root_path)
        self.x.to_df().to_csv(root_path / (name + '.csv'), index=False)

    def to_pickle(self, root_path, name):
        if not os.path.exists(root_path): os.makedirs(root_path)
        if not self.x: raise IOError('The dataset is empty, cannot pickle.')
        pickle.dump(self.x, open(Path(root_path) / (name + ".pickle"), "wb"), pickle.HIGHEST_PROTOCOL)


class MDPDataBunch(DataBunch):

    def close(self):
        if self.train_dl is not None: self.train_dl.env.close()
        if self.valid_dl is not None: self.valid_dl.env.close()

    @property
    def state_action_sample(self) -> Union[Tuple[State, Action], None]:
        ds = ifnone(self.train_ds, self.valid_ds)  # type: MDPDataset
        return ds.state, ds.action if ds is not None else None

    @classmethod
    def from_env(cls, env_name='CartPole-v1', max_steps=None, render='rgb_array', bs: int = 64,
                 feed_type=FEED_TYPE_STATE, num_workers: int = 0, memory_management_strategy='k_partitions_top',
                 split_env_init=True, device: torch.device = None, no_check: bool = False, x=None, val_x=None,
                 add_valid=True, **dl_kwargs):

        env = gym.make(env_name)
        memory_manager = partial(MDPMemoryManager, strategy=memory_management_strategy)
        train_list = MDPDataset(env, max_steps=max_steps, feed_type=feed_type, render=render, bs=bs,
                                memory_manager=memory_manager, x=x)
        if add_valid:
            valid_list = MDPDataset(env if split_env_init else gym.make(env_name), max_steps=max_steps, x=val_x,
                                    render=render, bs=bs,  feed_type=feed_type, memory_manager=memory_manager)
        else:
            valid_list = None
        path = './data/' + env_name + '_' + datetime.now().strftime('%Y%m%d%H%M%S')
        return cls.create(train_list, valid_list, path=path, num_workers=num_workers, bs=1, device=device, **dl_kwargs)

    @classmethod
    def from_pickle(cls, env_name=None, path: PathOrStr = None, **dl_kwargs):

        if path is None:
            path = [_ for _ in os.listdir('./data/') if _.__contains__(env_name.split('-v')[0].lower())]
            if not path: raise IOError(f'There is no pickle dirs file found in ./data/ with the env name {env_name}')
            path = Path('./data/' + path[0])

        path = Path(path)

        train_x = pickle.load(open(path / 'train.pickle', 'rb')) if (path / 'train.pickle').exists() else None
        val_x = pickle.load(open(path / 'valid.pickle', 'rb')) if (path / 'valid.pickle').exists() else None

        env_name = ifnone(env_name, Path(path).parts[-1].split('_')[0])
        return cls.from_env(env_name=env_name, x=train_x, val_x=val_x, **dl_kwargs)

    @classmethod
    def create(cls, train_ds: MDPDataset, valid_ds: MDPDataset = None, bs: int = 1, path='.',
               num_workers: int = defaults.cpus, device: torch.device = None, **dl_kwargs) -> 'DataBunch':
        """Create a `DataBunch` from `train_ds`, `valid_ds` and maybe `test_ds` with a batch size of `bs`.
        Passes `**dl_kwargs` to `DataLoader()`

        Since this is a MarkovProcess, the batches need to be `bs=1` (for now...)
        """
        datasets = cls._init_ds(train_ds, valid_ds, None)
        dls = [DataLoader(d, b, shuffle=s, drop_last=s, num_workers=num_workers, **dl_kwargs) for d, b, s in
               zip(datasets, (bs, bs, bs, bs), (False, False, False, False)) if d is not None]
        databunch = cls(path=path, device=device, *dls, **dl_kwargs)
        if valid_ds is None: databunch.valid_dl = None
        return databunch

    def to_csv(self):
        if self.train_ds is not None: self.train_ds.to_csv(self.path, 'train')
        if self.valid_ds is not None: self.valid_ds.to_csv(self.path, 'valid')

    def to_pickle(self, path=None):
        if self.train_ds is not None: self.train_ds.to_pickle(ifnone(path, self.path), 'train')
        if self.valid_ds is not None: self.valid_ds.to_pickle(ifnone(path, self.path), 'valid')

    @staticmethod
    def _init_ds(train_ds: Dataset, valid_ds: Dataset, test_ds: Optional[Dataset] = None):
        return [o for o in (train_ds, valid_ds, copy(train_ds), test_ds) if o is not None]


class MDPList(ItemList):
    _bunch = MDPDataBunch

    def __init__(self, items: Iterator, **kwargs):
        """
        Represents a MDP sequence over episodes.


        Notes:
            Two important fields for you to be aware of: `items` and `x`.
            `x` is just the v being used for directly being feed into the model.
            `items` contains an ndarray of MarkovDecisionProcessSliceAlpha instances. These contain the the primary v
            in x, but also the other important properties of a MDP.

        Args:
            items:
            feed_type:
            **kwargs:
        """
        # if items is not None:
        super().__init__(items, **kwargs)
        self.info = {}

    def set_recent_run_episode(self, episode):
        for i, item in enumerate(reversed(self.items)):
            if item.d and i != 0: break
            item.episode = episode
            self._update_info(episode, item)

    def clean(self, episode):
        self.info[episode][1] = True
        [item.clean() for item in self.items if item.episode == episode]

    def _update_info(self, ep, item: MDPStep):
        self.info[ep] = float(np.sum(self.info[ep][0] + float(item.reward))) if ep in self.info else float(item.reward)
        self.info[ep] = [self.info[ep], False]

    def add(self, items: 'ItemList'):
        [self._update_info(item.episode, item) for item in items.items]
        super().add(items)

    def to_df(self): return pd.DataFrame([i.obj for i in self.items])

    def to_dict(self): return [i.obj for i in self.items]

    def get(self, i): return self.items[i].data

    def reconstruct(self, t: Tensor, x: Tensor = None):
        raise NotImplementedError('Not sure when this will be important.')



import copy
from collections import deque
from math import ceil

import gym
from fastai.basic_train import *
from fastai.torch_core import *

from fast_rl.core.data_structures import SumTree


class ExplorationStrategy:
	def __init__(self, explore: bool = True):
		self.explore = explore

	def perturb(self, action, action_space):
		"""
		Base method just returns the action. Subclass, and change to return randomly / augmented actions.

		Should use `do_exploration` field. It is recommended that when you subclass / overload, you allow this field
		to completely bypass these actions.

		Args:
			action:
			action_space (gym.Space): The original gym space. Should contain information on the action type, and
			possible convenience methods for random action selection.

		Returns:

		"""
		_ = action_space
		return action

	def update(self, max_episodes, explore, **kwargs):
		self.explore = explore


class GreedyEpsilon(ExplorationStrategy):
	def __init__(self, epsilon_start, epsilon_end, decay, start_episode=0, end_episode=0, **kwargs):
		super().__init__(**kwargs)
		self.end_episode = end_episode
		self.start_episode = start_episode
		self.decay = decay
		self.epsilon_end = epsilon_end
		self.epsilon_start = epsilon_start
		self.epsilon = self.epsilon_start
		self.steps_done = 0

	def perturb(self, action, action_space: gym.Space):
		"""
		TODO for now does random discrete selection. Move to discrete soon.

		Args:
			action:
			action_space:

		Returns:
		"""
		if np.random.random() < self.epsilon and self.explore:
			return action_space.sample()
		else:
			return action

	def update(self, episode, end_episode=0, **kwargs):
		super(GreedyEpsilon, self).update(**kwargs)
		if self.explore:
			self.end_episode = end_episode
			self.epsilon = self.epsilon_end + (self.epsilon_start - self.epsilon_end) * \
						   math.exp(-1. * (self.steps_done * self.decay))
			self.steps_done += 1


class OrnsteinUhlenbeck(GreedyEpsilon):
	def __init__(self, size, mu=0., theta=0.15, sigma=0.2, **kwargs):
		"""

		References:
			[1] From https://math.stackexchange.com/questions/1287634/implementing-ornstein-uhlenbeck-in-matlab
			[2] Cumulatively based on

		Args:
			epsilon_start:
			epsilon_end:
			decay:
			**kwargs:
		"""
		super().__init__(**kwargs)
		self.sigma = sigma
		self.theta = theta
		self.mu = mu
		self.x = np.ones(size)

	def perturb(self, action, action_space):
		if self.explore:
			dx = self.theta * (self.mu - self.x) + self.sigma * np.array(
				[np.random.normal() for _ in range(len(self.x))])
		else:
			dx = np.zeros(self.x.shape)

		self.x += dx
		return self.epsilon * self.x + action


class Experience:
	def __init__(self, memory_size, reduce_ram=False):
		self.reduce_ram = reduce_ram
		self.max_size = memory_size
		self.callbacks = []

	@property
	def memory(self):
		return None

	def sample(self, **kwargs):
		pass

	def update(self, item, device, **kwargs):
		item.to(device=device)

	def refresh(self, **kwargs):
		pass


class ExperienceReplay(Experience):
	def __init__(self, memory_size, **kwargs):
		r"""
		Basic store-er of s space transitions for training agents.

		References:
			[1] Mnih, Volodymyr, et al. "Playing atari with deep reinforcement learning."
			arXiv preprint arXiv:1312.5602 (2013).

		Args:
			memory_size (int): Max N samples to store
		"""
		super().__init__(memory_size, **kwargs)
		self.max_size = memory_size
		self._memory = deque(maxlen=memory_size)

	@property
	def memory(self):
		return self._memory

	def __len__(self):
		return len(self._memory)

	def sample(self, batch, **kwargs):
		if len(self._memory) < batch: return self._memory
		return random.sample(self.memory, batch)

	def update(self, item, device, **kwargs):
		super().update(item, device, **kwargs)
		if self.reduce_ram: item.clean()
		self._memory.append(deepcopy(item))


class PriorityExperienceReplayCallback(LearnerCallback):
	def on_train_begin(self, **kwargs):
		self.learn.model.loss_func = partial(self.learn.model.memory.handle_loss,
											 base_function=self.learn.model.loss_func,
											 device=self.learn.data.device)


class PriorityExperienceReplay(Experience):

	def handle_loss(self, y, y_hat, base_function, device):
		return (base_function(y, y_hat) * torch.from_numpy(self.priority_weights).to(device=device).float()).mean()

	def __init__(self, memory_size, batch_size=64, epsilon=0.01, alpha=0.6, beta=0.4, b_inc=-0.001, **kwargs):
		"""
		Prioritizes sampling based on samples requiring the most learning.

		References:
			[1] Schaul, Tom, et al. "Prioritized experience replay." arXiv preprint arXiv:1511.05952 (2015).

		Args:
			batch_size (int): Size of sample, and thus size of expected index update.
			alpha (float): Changes the sampling behavior 1 (non-uniform) -> 0 (uniform)
			epsilon (float): Keeps the probabilities of items from being 0
			memory_size (int): Max N samples to store
		"""
		super().__init__(memory_size, **kwargs)
		self.batch_size = batch_size
		self.alpha = alpha
		self.beta = beta
		self.b_inc = b_inc
		self.priority_weights = None  # np.zeros(self.batch_size, dtype=float)
		self.epsilon = epsilon
		self.tree = SumTree(self.max_size)
		self.callbacks = [PriorityExperienceReplayCallback]
		# When sampled, store the sample indices for refresh.
		self._indices = None  # np.zeros(self.batch_size, dtype=int)

	@property
	def memory(self):
		return self.tree.data

	def __len__(self):
		return self.tree.n_entries

	def refresh(self, post_optimize, **kwargs):
		if post_optimize is not None:
			self.tree.update(self._indices.astype(int), np.abs(post_optimize['td_error']) + self.epsilon)

	def sample(self, batch, **kwargs):
		self.beta = np.min([1., self.beta + self.b_inc])
		# ranges = np.linspace(0, self.tree.total(), num=ceil(self.tree.total() / self.batch_size))
		ranges = np.linspace(0, ceil(self.tree.total() / batch), num=batch + 1)
		uniform_ranges = [np.random.uniform(ranges[i], ranges[i + 1]) for i in range(len(ranges) - 1)]
		try:
			self._indices, weights, samples = self.tree.batch_get(uniform_ranges)
		except ValueError:
			warn('Too few values to unpack. Your batch size is too small, when PER queries tree, all 0 values get'
				 ' ignored. We will retry until we can return at least one sample.')
			samples = self.sample(batch)

		self.priority_weights = self.tree.anneal_weights(weights, self.beta)
		return samples

	def update(self, item, device, **kwargs):
		"""
		Updates the tree of PER.

		Assigns maximal priority per [1] Alg:1, thus guaranteeing that sample being visited once.

		Args:
			item:

		Returns:

		"""
		super().update(item, device, **kwargs)
		maximal_priority = self.alpha
		if self.reduce_ram: item.clean()
		self.tree.add(np.abs(maximal_priority) + self.epsilon, item)


class HindsightExperienceReplay(Experience):
	def __init__(self, memory_size):
		"""

		References:
			[1] Andrychowicz, Marcin, et al. "Hindsight experience replay."
			Advances in Neural Information Processing Systems. 2017.

		Args:
			memory_size:
		"""
		super().__init__(memory_size)



import pytest


def pytest_addoption(parser):
    parser.addoption("--include_performance_tests", action="store_true",
                     help="Will run the performance tests which do full model testing. This could take a few"
                             "days to fully accomplish.")

@pytest.fixture()
def include_performance_tests(pytestconfig):
    return pytestconfig.getoption("include_performance_tests")


@pytest.fixture()
def skip_performance_check(include_performance_tests):
    if not include_performance_tests:
        pytest.skip('Skipping due to performance argument not specified. Add --include_performance_tests to not skip')



from functools import partial
from itertools import product

import pytest
from fast_rl.agents.ddpg_models import DDPGModule
from fastai.basic_train import torch, DatasetType

from fast_rl.agents.ddpg import create_ddpg_model, ddpg_learner
from fast_rl.core.agent_core import ExperienceReplay, PriorityExperienceReplay, OrnsteinUhlenbeck
from fast_rl.core.basic_train import AgentLearner
from fast_rl.core.data_block import FEED_TYPE_STATE, MDPDataBunch, FEED_TYPE_IMAGE
from fast_rl.core.metrics import RewardMetric, EpsilonMetric
from fast_rl.core.train import GroupAgentInterpretation, AgentInterpretation


p_model = [DDPGModule]
p_exp = [ExperienceReplay, PriorityExperienceReplay]
p_format = [FEED_TYPE_STATE]#, FEED_TYPE_IMAGE]
p_full_format = [FEED_TYPE_STATE]
p_envs = ['Walker2DPyBulletEnv-v0']

config_env_expectations = {
    'Walker2DPyBulletEnv-v0': {'action_shape': (1, 6), 'state_shape': (1, 22)}
}


@pytest.mark.parametrize(["model_cls", "s_format", "env"], list(product(p_model, p_format, p_envs)))
def test_ddpg_create_ddpg_model(model_cls, s_format, env):
    data = MDPDataBunch.from_env(env, render='rgb_array', bs=32, add_valid=False, feed_type=s_format)
    model = create_ddpg_model(data, model_cls)
    model.eval()
    model(data.state.s.float())

    assert config_env_expectations[env]['action_shape'] == (1, data.action.taken_action.shape[1])
    if s_format == FEED_TYPE_STATE:
        assert config_env_expectations[env]['state_shape'] == data.state.s.shape


@pytest.mark.parametrize(["model_cls", "s_format", "mem", "env"], list(product(p_model, p_format, p_exp, p_envs)))
def test_dddpg_ddpglearner(model_cls, s_format, mem, env):
    data = MDPDataBunch.from_env(env, render='rgb_array', bs=32, add_valid=False, feed_type=s_format)
    model = create_ddpg_model(data, model_cls)
    memory = mem(memory_size=1000, reduce_ram=True)
    exploration_method = OrnsteinUhlenbeck(size=data.action.taken_action.shape, epsilon_start=1, epsilon_end=0.1, decay=0.001)
    ddpg_learner(data=data, model=model, memory=memory, exploration_method=exploration_method)

    assert config_env_expectations[env]['action_shape'] == (1, data.action.taken_action.shape[1])
    if s_format == FEED_TYPE_STATE:
        assert config_env_expectations[env]['state_shape'] == data.state.s.shape


@pytest.mark.parametrize(["model_cls", "s_format", "mem", "env"], list(product(p_model, p_format, p_exp, p_envs)))
def test_ddpg_fit(model_cls, s_format, mem, env):
    data = MDPDataBunch.from_env(env, render='rgb_array', bs=10, max_steps=20, add_valid=False, feed_type=s_format)
    model = create_ddpg_model(data, model_cls, opt=torch.optim.RMSprop, layers=[20, 20])
    memory = mem(memory_size=100, reduce_ram=True)
    exploration_method = OrnsteinUhlenbeck(size=data.action.taken_action.shape, epsilon_start=1, epsilon_end=0.1, decay=0.001)
    learner = ddpg_learner(data=data, model=model, memory=memory, exploration_method=exploration_method)
    learner.fit(2)

    assert config_env_expectations[env]['action_shape'] == (1, data.action.taken_action.shape[1])
    if s_format == FEED_TYPE_STATE:
        assert config_env_expectations[env]['state_shape'] == data.state.s.shape

    del data
    del model
    del learner


@pytest.mark.usefixtures('skip_performance_check')
@pytest.mark.parametrize(["model_cls", "s_format", 'experience'],
                         list(product(p_model, p_format, p_exp)))
def test_ddpg_models_pendulum(model_cls, s_format, experience):
    group_interp = GroupAgentInterpretation()
    for i in range(5):
        print('\n')
        data = MDPDataBunch.from_env('Pendulum-v0', render='human', bs=64, add_valid=False, feed_type=s_format)
        exploration_method = OrnsteinUhlenbeck(size=data.action.taken_action.shape, epsilon_start=1, epsilon_end=0.1,
                                               decay=0.001)
        memory = experience(memory_size=1000000, reduce_ram=True)
        model = create_ddpg_model(data=data, base_arch=model_cls)
        learner = ddpg_learner(data=data, model=model, memory=memory, exploration_method=exploration_method)
        learner.fit(450)

        meta = f'{experience.__name__}_{"FEED_TYPE_STATE" if s_format == FEED_TYPE_STATE else "FEED_TYPE_IMAGE"}'
        interp = AgentInterpretation(learner, ds_type=DatasetType.Train)
        interp.plot_rewards(cumulative=True, per_episode=True, group_name=meta)
        group_interp.add_interpretation(interp)
        group_interp.to_pickle(f'../docs_src/data/pendulum_{model.name.lower()}/', f'{model.name.lower()}_{meta}')

        del learner
        del model
        del data


@pytest.mark.usefixtures('skip_performance_check')
@pytest.mark.parametrize(["model_cls", "s_format", 'experience'],
                         list(product(p_model, p_full_format, p_exp)))
def test_ddpg_models_mountain_car_continuous(model_cls, s_format, experience):
    group_interp = GroupAgentInterpretation()
    for i in range(5):
        print('\n')
        data = MDPDataBunch.from_env('MountainCarContinuous-v0', render='human', bs=40, add_valid=False, feed_type=s_format)
        exploration_method = OrnsteinUhlenbeck(size=data.action.taken_action.shape, epsilon_start=1, epsilon_end=0.1,
                                               decay=0.0001)
        memory = experience(memory_size=1000000, reduce_ram=True)
        model = create_ddpg_model(data=data, base_arch=model_cls)
        learner = ddpg_learner(data=data, model=model, memory=memory, exploration_method=exploration_method)
        learner.fit(450)

        meta = f'{experience.__name__}_{"FEED_TYPE_STATE" if s_format == FEED_TYPE_STATE else "FEED_TYPE_IMAGE"}'
        interp = AgentInterpretation(learner, ds_type=DatasetType.Train)
        interp.plot_rewards(cumulative=True, per_episode=True, group_name=meta)
        group_interp.add_interpretation(interp)
        group_interp.to_pickle(f'../docs_src/data/mountaincarcontinuous_{model.name.lower()}/',
                               f'{model.name.lower()}_{meta}')

        del learner
        del model
        del data


@pytest.mark.usefixtures('skip_performance_check')
@pytest.mark.parametrize(["model_cls", "s_format", 'experience'],
                         list(product(p_model, p_full_format, p_exp)))
def test_ddpg_models_reach(model_cls, s_format, experience):
    group_interp = GroupAgentInterpretation()
    for i in range(5):
        print('\n')
        data = MDPDataBunch.from_env('ReacherPyBulletEnv-v0', render='human', bs=40, add_valid=False,feed_type=s_format)
        exploration_method = OrnsteinUhlenbeck(size=data.action.taken_action.shape, epsilon_start=1, epsilon_end=0.1,
                                               decay=0.0001)
        memory = experience(memory_size=1000000, reduce_ram=True)
        model = create_ddpg_model(data=data, base_arch=model_cls)
        learner = ddpg_learner(data=data, model=model, memory=memory, exploration_method=exploration_method)
        learner.fit(450)

        meta = f'{experience.__name__}_{"FEED_TYPE_STATE" if s_format == FEED_TYPE_STATE else "FEED_TYPE_IMAGE"}'
        interp = AgentInterpretation(learner, ds_type=DatasetType.Train)
        interp.plot_rewards(cumulative=True, per_episode=True, group_name=meta)
        group_interp.add_interpretation(interp)
        group_interp.to_pickle(f'../docs_src/data/reacher_{model.name.lower()}/',
                               f'{model.name.lower()}_{meta}')

        del learner
        del model
        del data


@pytest.mark.usefixtures('skip_performance_check')
@pytest.mark.parametrize(["model_cls", "s_format", 'experience'],
                         list(product(p_model, p_full_format, p_exp)))
def test_ddpg_models_walker(model_cls, s_format, experience):
    group_interp = GroupAgentInterpretation()
    for i in range(5):
        print('\n')
        data = MDPDataBunch.from_env('Walker2DPyBulletEnv-v0', render='human', bs=64, add_valid=False,
                                     feed_type=s_format)
        exploration_method = OrnsteinUhlenbeck(size=data.action.taken_action.shape, epsilon_start=1, epsilon_end=0.1,
                                               decay=0.0001)
        memory = experience(memory_size=1000000, reduce_ram=True)
        model = create_ddpg_model(data=data, base_arch=model_cls)
        learner = ddpg_learner(data=data, model=model, memory=memory, exploration_method=exploration_method)
        learner.fit(2000)

        meta = f'{experience.__name__}_{"FEED_TYPE_STATE" if s_format == FEED_TYPE_STATE else "FEED_TYPE_IMAGE"}'
        interp = AgentInterpretation(learner, ds_type=DatasetType.Train)
        interp.plot_rewards(cumulative=True, per_episode=True, group_name=meta)
        group_interp.add_interpretation(interp)
        group_interp.to_pickle(f'../docs_src/data/walker2d_{model.name.lower()}/',
                               f'{model.name.lower()}_{meta}')

        del learner
        del model
        del data


@pytest.mark.usefixtures('skip_performance_check')
@pytest.mark.parametrize(["model_cls", "s_format", 'experience'],
                         list(product(p_model, p_full_format, p_exp)))
def test_ddpg_models_ant(model_cls, s_format, experience):
    group_interp = GroupAgentInterpretation()
    for i in range(5):
        print('\n')
        data = MDPDataBunch.from_env('AntPyBulletEnv-v0', render='human', bs=64, add_valid=False,feed_type=s_format)
        exploration_method = OrnsteinUhlenbeck(size=data.action.taken_action.shape, epsilon_start=1, epsilon_end=0.1,
                                               decay=0.00001)
        memory = experience(memory_size=1000000, reduce_ram=True)
        model = create_ddpg_model(data=data, base_arch=model_cls, lr=1e-3, actor_lr=1e-4,)
        learner = ddpg_learner(data=data, model=model, memory=memory, exploration_method=exploration_method,
                               opt_func=torch.optim.Adam, callback_fns=[RewardMetric, EpsilonMetric])
        learner.fit(1000)

        meta = f'{experience.__name__}_{"FEED_TYPE_STATE" if s_format == FEED_TYPE_STATE else "FEED_TYPE_IMAGE"}'
        interp = AgentInterpretation(learner, ds_type=DatasetType.Train)
        interp.plot_rewards(cumulative=True, per_episode=True, group_name=meta)
        group_interp.add_interpretation(interp)
        group_interp.to_pickle(f'../docs_src/data/ant_{model.name.lower()}/',
                               f'{model.name.lower()}_{meta}')

        del learner
        del model
        del data


@pytest.mark.usefixtures('skip_performance_check')
@pytest.mark.parametrize(["model_cls", "s_format", 'experience'],
                         list(product(p_model, p_full_format, p_exp)))
def test_ddpg_models_halfcheetah(model_cls, s_format, experience):
    group_interp = GroupAgentInterpretation()
    for i in range(5):
        print('\n')
        data = MDPDataBunch.from_env('HalfCheetahPyBulletEnv-v0', render='human', bs=64, add_valid=False,
                                     feed_type=s_format)
        exploration_method = OrnsteinUhlenbeck(size=data.action.taken_action.shape, epsilon_start=1, epsilon_end=0.1,
                                               decay=0.0001)
        memory = experience(memory_size=1000000, reduce_ram=True)
        model = create_ddpg_model(data=data, base_arch=model_cls)
        learner = ddpg_learner(data=data, model=model, memory=memory, exploration_method=exploration_method)
        learner.fit(2000)

        meta = f'{experience.__name__}_{"FEED_TYPE_STATE" if s_format == FEED_TYPE_STATE else "FEED_TYPE_IMAGE"}'
        interp = AgentInterpretation(learner, ds_type=DatasetType.Train)
        interp.plot_rewards(cumulative=True, per_episode=True, group_name=meta)
        group_interp.add_interpretation(interp)
        group_interp.to_pickle(f'../docs_src/data/halfcheetah_{model.name.lower()}/',
                               f'{model.name.lower()}_{meta}')

        del learner
        del model
        del data


from fast_rl.core.data_structures import print_tree, SumTree


def test_sum_tree_with_max_size():
    memory = SumTree(10)

    values = [1, 1, 1, 1, 1, 1]
    data = [f'data with priority: {i}' for i in values]

    for element, value in zip(data, values):
        memory.add(value, element)

    print_tree(memory)





from multiprocessing.pool import Pool

from fastai.basic_train import Learner, warn, ifnone, F, List


class WrapperLossFunc(object):
    def __init__(self, learn):
        self.learn = learn

    def __call__(self, *args, **kwargs):
        return self.learn.model.loss


class AgentLearner(Learner):

    def __init__(self, data, loss_func=None, callback_fns=None, **kwargs):
        super().__init__(data=data, callback_fns=ifnone(callback_fns, []) + data.callback, **kwargs)
        self.model.loss_func = ifnone(loss_func, F.mse_loss)
        self.loss_func = None
        self._loss_func = WrapperLossFunc(self)

    @property
    def warming_up(self):
        return self.data.bs > len(self.data.x)

    def init_loss_func(self):
        r"""
         Initializes the loss function wrapper for logging loss.

         Since most RL models have a period of warming up such as filling tree buffers, we cannot log any loss.
         By default, the learner will have a `None` loss function, and so the fit function will not try to log that
         loss.
         """
        self.loss_func = self._loss_func

    def interpret_q(self, xi):
        raise NotImplemented


class PipeLine(object):
    def __init__(self, n_threads, pipe_line_function):
        warn(Warning('Currently not super useful. Seems to have issues with running a single env in multiple threads.'))
        self.pipe_line_function = pipe_line_function
        self.n_threads = n_threads
        self.pool = Pool(self.n_threads)

    def start(self, n_runs):
        return self.pool.map(self.pipe_line_function,  range(n_runs))


from math import ceil

from fastai.layers import Flatten
from fastai.tabular import TabularModel, OptimWrapper
from fastai.torch_core import *


def init_cnn(mod):
    if getattr(mod, 'bias', None) is not None: nn.init.constant_(mod.bias, 0)
    if isinstance(mod, (nn.Conv2d, nn.Linear)): nn.init.kaiming_normal_(mod.weight)
    for sub_mod in mod.children(): init_cnn(sub_mod)


def ks_stride(ks, stride, w, h, n_blocks, kern_proportion=.1, stride_proportion=0.3):
    kernels, strides, max_dim = [], [], max((w, h))
    for i in range(len(n_blocks)):
        kernels.append(max_dim * kern_proportion)
        strides.append(kernels[-1] * stride_proportion)
        max_dim = (max_dim - kernels[-1]) / strides[-1]
        assert max_dim > 1

    return ifnone(ks, map(ceil, kernels)), ifnone(stride, map(ceil, strides))


def conv_bn_lrelu(ni: int, nf: int, ks: int = 3, stride: int = 1, pad=True) -> nn.Sequential:
    r""" Create a sequence Conv2d->BatchNorm2d->LeakyReLu layer. (from darknet.py) """
    return nn.Sequential(
        nn.Conv2d(ni, nf, kernel_size=ks, bias=False, stride=stride, padding=(ks // 2) if pad else 0),
        nn.BatchNorm2d(nf),
        nn.LeakyReLU(negative_slope=0.1, inplace=True))


class ChannelTranspose(Module):
    def forward(self, xi: Tensor):
        return xi.transpose(3, 1).transpose(3, 2)


class TabularEmbedWrapper(Module):
    def __init__(self, tabular_model: Module):
        super().__init__()
        self.tabular_model = tabular_model

    def forward(self, xi: Tensor, *args):
        return self.tabular_model(xi, xi)


class DQNModule(Module):

    def __init__(self, ni: int, ao: int, layers: Collection[int], discount: float = 0.99, lr=0.001,
                 n_conv_blocks: Collection[int] = 0, nc=3, opt=None, emb_szs: ListSizes = None, loss_func=None,
                 w=-1, h=-1, ks: Union[None, list]=None, stride: Union[None, list]=None, grad_clip=1,
                 conv_kern_proportion=0.1, stride_proportion=0.1, pad=False):
        r"""
        Basic DQN Module.

        Args:
            ni: Number of inputs. Expecting a flat state `[1 x ni]`
            ao: Number of actions to output.
            layers: Number of layers where is determined per element.
            n_conv_blocks: If `n_conv_blocks` is not 0, then convolutional blocks will be added
                           to the head on top of existing linear layers.
            nc: Number of channels that will be expected by the convolutional blocks.
        """
        super().__init__()
        self.name = 'DQN'
        self.loss_func = loss_func
        self.discount = discount
        self.gradient_clipping_norm = grad_clip
        self.lr = lr
        self.switched = False
        self.ks, self.stride = ([], []) if len(n_conv_blocks) == 0 else ks_stride(ks, stride, w, h, n_conv_blocks, conv_kern_proportion, stride_proportion)
        self.action_model = nn.Sequential()
        _layers = [conv_bn_lrelu(nc, self.nf, ks=ks, stride=stride, pad=pad) for self.nf, ks, stride in zip(n_conv_blocks, self.ks, self.stride)]

        if _layers: ni = self.setup_conv_block(_layers=_layers, ni=ni, nc=nc, w=w, h=h)
        self.setup_linear_block(_layers=_layers, ni=ni, nc=nc, w=w, h=h, emb_szs=emb_szs, layers=layers, ao=ao)
        self.init_weights(self.action_model)
        self.opt = OptimWrapper.create(ifnone(optim.Adam, opt), lr=self.lr, layer_groups=[self.action_model])

    def setup_conv_block(self, _layers, ni, nc, w, h):
        self.action_model.add_module('conv_block', nn.Sequential(*(self.fix_switched_channels(ni, nc, _layers) + [Flatten()])))
        training = self.action_model.training
        self.action_model.eval()
        ni = int(self.action_model(torch.zeros((1, w, h, nc) if self.switched else (1, nc, w, h))).view(-1, ).shape[0])
        self.action_model.train(training)
        return ni

    def setup_linear_block(self, _layers, ni, nc, w, h, emb_szs, layers, ao):
        tabular_model = TabularModel(emb_szs=emb_szs, n_cont=ni if not emb_szs else 0, layers=layers, out_sz=ao)
        if not emb_szs: tabular_model.embeds = None
        self.action_model.add_module('lin_block', TabularEmbedWrapper(tabular_model))

    def fix_switched_channels(self, current_channels, expected_channels, layers: list):
        if current_channels == expected_channels:
            return layers
        else:
            self.switched = True
            return [ChannelTranspose()] + layers

    def forward(self, xi: Tensor):
        training = self.training
        if xi.shape[0] == 1: self.eval()
        pred = self.action_model(xi)
        if training: self.train()
        return pred

    def init_weights(self, m):
        if type(m) == nn.Linear:
            torch.nn.init.xavier_uniform_(m.weight)
            m.bias.data.fill_(0.01)

    def sample_mask(self, d):
        return torch.sub(1.0, d)

    def optimize(self, sampled):
        r"""Uses ER to optimize the Q-net (without fixed targets).

        Uses the equation:

        .. math::
                Q^{*}(s, a) = \mathbb{E}_{s'∼ \Big\epsilon} \Big[r + \lambda \displaystyle\max_{a'}(Q^{*}(s' , a'))
                \;|\; s, a \Big]


        Returns (dict): Optimization information

        """
        r = torch.cat([item.reward.float() for item in sampled])#.to(self.data.device)
        s_prime = torch.cat([item.s_prime for item in sampled])#.to(self.data.device)
        s = torch.cat([item.s for item in sampled])#.to(self.data.device)
        a = torch.cat([item.a.long() for item in sampled])#.to(self.data.device)
        d = torch.cat([item.done.float() for item in sampled])#.to(self.data.device)
        masking = self.sample_mask(d)

        y_hat = self.y_hat(s, a)
        y = self.y(s_prime, masking, r, y_hat)

        loss = self.loss_func(y, y_hat)

        if self.training:
            self.opt.zero_grad()
            loss.backward()
            torch.nn.utils.clip_grad_norm_(self.action_model.parameters(), self.gradient_clipping_norm)
            for param in self.action_model.parameters():
                if param.grad is not None: param.grad.data.clamp_(-1, 1)
            self.opt.step()

        with torch.no_grad():
            self.loss = loss
            post_info = {'td_error': to_detach(y - y_hat).cpu().numpy()}
            return post_info

    def y_hat(self, s, a):
        return self.action_model(s).gather(1, a)

    def y(self, s_prime, masking, r, y_hat):
        return self.discount * self.action_model(s_prime).max(1)[0].unsqueeze(1) * masking + r.expand_as(y_hat)


class FixedTargetDQNModule(DQNModule):
    def __init__(self, ni: int, ao: int, layers: Collection[int], tau=1, **kwargs):
        super().__init__(ni, ao, layers, **kwargs)
        self.name = 'Fixed Target DQN'
        self.tau = tau
        self.target_model = copy(self.action_model)

    def target_copy_over(self):
        r""" Updates the target network from calls in the FixedTargetDQNTrainer callback."""
        # self.target_net.load_state_dict(self.action_model.state_dict())
        for target_param, local_param in zip(self.target_model.parameters(), self.action_model.parameters()):
            target_param.data.copy_(self.tau * local_param.data + (1.0 - self.tau) * target_param.data)

    def y(self, s_prime, masking, r, y_hat):
        r"""
        Uses the equation:

        .. math::

                Q^{*}(s, a) = \mathbb{E}_{s'∼ \Big\epsilon} \Big[r + \lambda \displaystyle\max_{a'}(Q^{*}(s' , a'))
                \;|\; s, a \Big]

        """
        return self.discount * self.target_model(s_prime).max(1)[0].unsqueeze(1) * masking + r.expand_as(y_hat)


class DoubleDQNModule(FixedTargetDQNModule):
    def __init__(self, ni: int, ao: int, layers: Collection[int], **kwargs):
        super().__init__(ni, ao, layers, **kwargs)
        self.name = 'DDQN'

    def calc_y(self, s_prime, masking, r, y_hat):
        return self.discount * self.target_model(s_prime).gather(1, self.action_model(s_prime).argmax(1).unsqueeze(
            1)) * masking + r.expand_as(y_hat)


class DuelingBlock(nn.Module):
    def __init__(self, ao, stream_input_size):
        super().__init__()

        self.val = nn.Linear(stream_input_size, 1)
        self.adv = nn.Linear(stream_input_size, ao)

    def forward(self, xi):
        r"""Splits the base neural net output into 2 streams to evaluate the advantage and v of the s space and
        corresponding actions.

        .. math::
           Q(s,a;\; \Theta, \\alpha, \\beta) = V(s;\; \Theta, \\beta) + A(s, a;\; \Theta, \\alpha) - \\frac{1}{|A|}
           \\Big\\sum_{a'} A(s, a';\; \Theta, \\alpha)

        """
        val, adv = self.val(xi), self.adv(xi)
        xi = val.expand_as(adv) + (adv - adv.mean()).squeeze(0)
        return xi


class DuelingDQNModule(FixedTargetDQNModule):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = 'Dueling DQN'

    def setup_linear_block(self, _layers, ni, nc, w, h, emb_szs, layers, ao):
        model = TabularModel(emb_szs=emb_szs, n_cont=ni if not emb_szs else 0, layers=layers, out_sz=ao, use_bn=False)
        if not emb_szs: model.embeds = None
        model.layers, removed_layer = split_model(model.layers, [last_layer(model)])
        ni = removed_layer[0].in_features
        self.action_model.add_module('lin_block', TabularEmbedWrapper(model))
        self.action_model.add_module('dueling_block', DuelingBlock(ao, ni))


class DoubleDuelingModule(DuelingDQNModule, DoubleDQNModule):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = 'DDDQN'



# from itertools import product
#
# import pytest
# from fastai.basic_data import DatasetType
#
# from fast_rl.agents.dqn import DQN
# from fast_rl.core.data_block import MDPDataBunch
# from fast_rl.core.basic_train import AgentLearner
# from fast_rl.core.train import AgentInterpretation, GroupAgentInterpretation
#
#
# reward_plot_list = list(product([True, False], repeat=2))
#
# def test_interpretation_reward_group_plot():
#     group_interp = GroupAgentInterpretation()
#     for i in range(2):
#         data = MDPDataBunch.from_env('CartPole-v0', render='rgb_array', bs=4, add_valid=False)
#         model = DQN(data)
#         learn = AgentLearner(data, model)
#         learn.fit(2)
#
#         interp = AgentInterpretation(learn=learn, ds_type=DatasetType.Train)
#         interp.plot_rewards(cumulative=True, per_episode=True, group_name='run1')
#         group_interp.add_interpretation(interp)
#
#     group_interp.plot_reward_bounds(return_fig=True, per_episode=True).show()
#
# @pytest.mark.parametrize(["per_episode", "cumulative"], reward_plot_list)
# def test_interpretation_reward_plot(per_episode, cumulative):
#     data = MDPDataBunch.from_env('CartPole-v0', render='rgb_array')
#     model = DQN(data)
#     learn = AgentLearner(data, model)
#     learn.fit(2)
#
#     interp = AgentInterpretation(learn=learn)
#     interp.plot_rewards(return_fig=True, per_episode=per_episode, cumulative=cumulative).show()
#     data.train_ds.env.close()
#     data.valid_ds.env.close()



import pytest

from fast_rl.agents.dqn import *
from fast_rl.agents.dqn_models import FixedTargetDQNModule
from fast_rl.core.agent_core import *
from fast_rl.core.data_block import *
from fast_rl.core.train import *

p_model = [FixedTargetDQNModule]
p_exp = [ExperienceReplay]
p_format = [FEED_TYPE_STATE]

config_env_expectations = {
	'CartPole-v1': {'action_shape': (1, 2), 'state_shape': (1, 4)},
	'maze-random-5x5-v0': {'action_shape': (1, 4), 'state_shape': (1, 2)}
}


@pytest.mark.parametrize(["model_cls", "s_format", "mem"], list(product(p_model, p_format, p_exp)))
def test_train_gym_maze_interpretation(model_cls, s_format, mem):
	success = False
	while not success:
		try:
			data = MDPDataBunch.from_env('maze-random-5x5-v0', render='rgb_array', bs=5, max_steps=50,
										 add_valid=False, feed_type=s_format)
			model = create_dqn_model(data, model_cls, opt=torch.optim.RMSprop)
			memory = mem(10000)
			exploration_method = GreedyEpsilon(epsilon_start=1, epsilon_end=0.1, decay=0.001)
			learner = dqn_learner(data=data, model=model, memory=memory, exploration_method=exploration_method)
			learner.fit(1)

			interp = GymMazeInterpretation(learner, ds_type=DatasetType.Train)
			for i in range(-1, 4): interp.plot_heat_map(action=i)

			success = True
		except Exception as e:
			if not str(e).__contains__('Surface'):
				raise Exception


@pytest.mark.parametrize(["model_cls", "s_format", "mem"], list(product(p_model, p_format, p_exp)))
def test_train_q_value_interpretation(model_cls, s_format, mem):
	success = False
	while not success:
		try:
			data = MDPDataBunch.from_env('maze-random-5x5-v0', render='rgb_array', bs=5, max_steps=50,
										 add_valid=False, feed_type=s_format)
			model = create_dqn_model(data, model_cls, opt=torch.optim.RMSprop)
			memory = mem(10000)
			exploration_method = GreedyEpsilon(epsilon_start=1, epsilon_end=0.1, decay=0.001)
			learner = dqn_learner(data=data, model=model, memory=memory, exploration_method=exploration_method)
			learner.fit(1)

			interp = QValueInterpretation(learner, ds_type=DatasetType.Train)
			interp.plot_q()

			success = True
		except Exception as e:
			if not str(e).__contains__('Surface'):
				raise Exception(e)

#
# def test_groupagentinterpretation_from_pickle():
#     group_interp = GroupAgentInterpretation.from_pickle('./data/cartpole_dqn',
#                                                         'dqn_PriorityExperienceReplay_FEED_TYPE_STATE')
#     group_interp.plot_reward_bounds(return_fig=True, per_episode=True, smooth_groups=5).show()
#
#
# def test_groupagentinterpretation_analysis():
#     group_interp = GroupAgentInterpretation.from_pickle('./data/cartpole_dqn',
#                                                         'dqn_PriorityExperienceReplay_FEED_TYPE_STATE')
#     assert isinstance(group_interp.analysis, list)
#     group_interp.in_notebook = True
#     assert isinstance(group_interp.analysis, pd.DataFrame)




#
# def test_interpretation_reward_group_plot():
#     group_interp = GroupAgentInterpretation()
#     group_interp2 = GroupAgentInterpretation()
#
#     for i in range(2):
#         data = MDPDataBunch.from_env('CartPole-v0', render='rgb_array', bs=4, add_valid=False)
#         model = DQN(data)
#         learn = AgentLearner(data, model)
#         learn.fit(2)
#
#         interp = AgentInterpretation(learn=learn, ds_type=DatasetType.Train)
#         interp.plot_rewards(cumulative=True, per_episode=True, group_name='run1')
#         group_interp.add_interpretation(interp)
#         group_interp2.add_interpretation(interp)
#
#     group_interp.plot_reward_bounds(return_fig=True, per_episode=True).show()
#     group_interp2.plot_reward_bounds(return_fig=True, per_episode=True).show()
#
#     new_interp = group_interp.merge(group_interp2)
#     assert len(new_interp.groups) == len(group_interp.groups) + len(group_interp2.groups), 'Lengths do not match'
#     new_interp.plot_reward_bounds(return_fig=True, per_episode=True).show()


import gym
from gym import error


class b_colors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


def list_in_str(s: str, str_list: list, make_lower=True):
    if make_lower: s = s.lower()
    return any([s.__contains__(el) for el in str_list])


def is_goal_env(env, suppress_errors=True):
    msg = 'GoalEnv requires the "{}" key to be part of the observation d.'
    # Enforce that each GoalEnv uses a Goal-compatible observation space.
    if not isinstance(env.observation_space, gym.spaces.Dict):
        if not suppress_errors:
            raise error.Error('GoalEnv requires an observation space of type gym.spaces.Dict')
        else:
            return False
    for key in ['observation', 'achieved_goal', 'desired_goal']:
        if key not in env.observation_space.spaces:
            if not suppress_errors:
                raise error.Error(msg.format(key))
            else:
                return False
    return True



import pytest
from fastai.basic_train import ItemLists

def validate_item_list(item_list: ItemLists):
    # Check items
    for i, item in enumerate(item_list.items):
        if item.done: assert not item_list.items[
            i - 1].done, f'The dataset has duplicate "done\'s" that are consecutive.'
        assert item.state.s is not None, f'The item: {item}\'s state is None'
        assert item.state.s_prime is not None, f'The item: {item}\'s state prime is None'



# @pytest.mark.parametrize("env", sorted(['CartPole-v0']))
# def test_mdp_from_pickle(env):
#     data = MDPDataBunch.from_env(env, render='rgb_array')
#     model = DQN(data)
#     learner = AgentLearner(data, model)
#     learner.fit(2)
#     data.to_pickle(path='data/CartPole-v0_testing')
#     data = MDPDataBunch.from_pickle(path='data/CartPole-v0_testing')
#     del data
#
#
# @pytest.mark.parametrize("env", sorted(['CartPole-v0']))
# def test_mdp_to_csv(env):
#     data = MDPDataBunch.from_env(env, render='rgb_array')
#     model = DQN(data)
#     learner = AgentLearner(data, model)
#     learner.fit(2)
#     data.to_csv()
#     data.train_ds.env.close()
#     data.valid_ds.env.close()
#     del learner
#
#
# @pytest.mark.parametrize("env", sorted(['CartPole-v0']))
# def test_mdp_to_pickle(env):
#     data = MDPDataBunch.from_env(env, render='rgb_array')
#     model = DQN(data)
#     learner = AgentLearner(data, model)
#     learner.fit(2)
#     data.to_pickle()
#     data.train_ds.env.close()
#     data.valid_ds.env.close()
#     del learner
#
#
# @pytest.mark.parametrize("env", sorted(['CartPole-v0']))
# def test_mdp_clean_callback(env):
#     data = MDPDataBunch.from_env(env, render='rgb_array')
#     model = DQN(data)
#     learner = AgentLearner(data, model)
#     learner.fit(15)
#     data.train_ds.env.close()
#     data.valid_ds.env.close()
#     del learner
#
#
# @pytest.mark.parametrize("env", sorted(['CartPole-v0']))
# def test_mdp_databunch(env):
#     data = MDPDataBunch.from_env(env, add_valid=False, render='rgb_array')
#     for i in range(5):
#         for _ in data.train_ds:
#             data.train_ds.action = Action(taken_action=data.train_ds.action.action_space.sample(),
#                                           action_space=data.train_ds.action.action_space)
#
#     validate_item_list(data.train_ds.x)
#     del data
#
#
# @pytest.mark.parametrize("env", sorted(['CartPole-v0']))
# def test_mdp_dataset_iter(env):
#     dataset = MDPDataset(gym.make(env), memory_manager = partial(MDPMemoryManager, strategy='k_partition_top'), bs=8,
#                          render='rgb_array')
#
#     for epoch in range(5):
#         for el in dataset:
#             dataset.action.set_single_action(dataset.env.action_space.sample())
#
#     # Check items
#     for i, item in enumerate(dataset.x.items):
#         if item.done: assert not dataset.x.items[
#             i - 1].done, f'The dataset has duplicate "done\'s" that are consecutive.'
#         assert item.state.s is not None, f'The item: {item}\'s state is None'
#         assert item.state.s_prime is not None, f'The item: {item}\'s state prime is None'
#     del dataset
#
#
# @pytest.mark.parametrize("env", sorted(['CartPole-v0']))
# def test_mdpdataset_init(env):
#     try:
#         init_env = gym.make(env)
#     except error.DependencyNotInstalled as e:
#         print(e)
#         return
#
#     data = MDPDataset(init_env, None, 64, 'rgb_array')
#
#     try:
#         max_steps = data.max_steps
#         assert max_steps is not None, f'Max steps is None for env {env}'
#     except MaxEpisodeStepsMissingError as e:
#         return
#
#     envs_to_test = {
#         'CartPole-v0': 200,
#         'MountainCar-v0': 200,
#         'maze-v0': 2000
#     }
#
#     if env in envs_to_test:
#         assert envs_to_test[env] == max_steps, f'Env {env} is the wrong step amount'
#     del data
#
#
# @pytest.mark.parametrize("env", sorted(['CartPole-v0']))
# def test_bound_init(env):
#     try:
#         init_env = gym.make(env)
#     except error.DependencyNotInstalled as e:
#         print(e)
#         return
#
#     for bound in (Bounds(init_env.action_space), Bounds(init_env.observation_space)):
#         if env.lower().__contains__('continuous'):
#             assert bound.n_possible_values == np.inf, f'Env {env} is continuous, should have inf v.'
#         if env.lower().__contains__('deterministic'):
#             assert bound.n_possible_values != np.inf, f'Env {env} is deterministic, should have discrete v.'
#     init_env.close()
#
# @pytest.mark.parametrize("env", sorted(['CartPole-v0']))
# def test_action_init(env):
#     try:
#         init_env = gym.make(env)
#     except error.DependencyNotInstalled as e:
#         print(e)
#         return
#
#     taken_action = init_env.action_space.sample()
#     raw_action = np.random.rand(len(Bounds(init_env.action_space).max))
#     init_env.reset()
#     _ = init_env.step(taken_action)
#
#     action = Action(taken_action=taken_action, raw_action=raw_action, action_space=init_env.action_space)
#
#     if list_in_str(env, ['mountaincar-', 'cartpole', 'pong']):
#         assert any([action.taken_action.dtype in (int, torch.int, torch.int64)]), f'Action is wrong dtype {action}'
#         assert any([action.raw_action.dtype in (float, torch.float32, torch.float64)]), f'Action is wrong dtype {action}'
#     if list_in_str(env, ['carracing', 'pendulum']):
#         assert any([action.taken_action.dtype in (float, torch.float32, torch.float64)]), f'Action is wrong dtype {action}'
#         assert any([action.raw_action.dtype in (float, torch.float32, torch.float64)]), f'Action is wrong dtype {action}'
#     init_env.close()
#
#
# @pytest.mark.parametrize("env", sorted(['CartPole-v0']))
# def test_state_init(env):
#     try:
#         init_env = gym.make(env)
#     except error.DependencyNotInstalled as e:
#         print(e)
#         return
#
#     taken_action = init_env.action_space.sample()
#     state = init_env.reset()
#     state_prime, reward, done, info = init_env.step(taken_action)
#     State(state, state_prime, init_env.observation_space)
#     init_env.close()
#
#
# @pytest.mark.parametrize("env", sorted(['CartPole-v0']))
# def test_state_str(env):
#     try:
#         init_env = gym.make(env)
#     except error.DependencyNotInstalled as e:
#         print(e)
#         return
#
#     render = 'rgb_array'
#     if isinstance(init_env, TimeLimit) and isinstance(init_env.unwrapped, (AlgorithmicEnv, discrete.DiscreteEnv)):
#         render = 'ansi' if render == 'rgb_array' else render
#
#     taken_action = init_env.action_space.sample()
#     state = init_env.reset()
#
#     try:
#         alt_s = init_env.render(render)
#     except NotImplementedError:
#         return
#
#     state_prime, reward, done, info = init_env.step(taken_action)
#     alt_s_prime = init_env.render(render)
#     State(state, state_prime, alt_s, alt_s_prime, init_env.observation_space).__str__()
#     init_env.close()
#
#
# # @pytest.mark.parametrize("env", sorted(['CartPole-v0', 'maze-random-5x5-v0']))
# # def test_state_full_episode(env):
# #     try:
# #         init_env = gym.make(env)
# #     except error.DependencyNotInstalled as e:
# #         print(e)
# #         return
# #
# #     done = False
# #     state = init_env.reset()
# #     while not done:
# #         taken_action = init_env.action_space.sample()
# #         alt_state = init_env.render('rgb_array')
# #         state_prime, reward, done, info = init_env.step(taken_action)
# #         alt_s_prime = init_env.render('rgb_array')
# #         State(state, state_prime, alt_state, alt_s_prime, init_env.observation_space)
# #         state = state_prime
# #         if done:
# #             assert state_prime is not None, 'State prime is None, this should not have happened.'
# #     init_env.close()



import torch
from fastai.basic_train import LearnerCallback, Any
from fastai.callback import Callback, is_listy, add_metrics


class EpsilonMetric(LearnerCallback):
    _order = -20  # Needs to run before the recorder

    def __init__(self, learn):
        super().__init__(learn)
        self.epsilon = 0
        if not hasattr(self.learn, 'exploration_method'):
            raise ValueError('Your model is not using an exploration strategy! Please use epsilon based exploration')
        if not hasattr(self.learn.exploration_method, 'epsilon'):
            raise ValueError('Please use epsilon based exploration (should have an epsilon field)')

    # noinspection PyUnresolvedReferences
    def on_train_begin(self, **kwargs):
        self.learn.recorder.add_metric_names(['epsilon'])

    def on_epoch_end(self, last_metrics, **kwargs):
        self.epsilon = self.learn.exploration_method.epsilon
        if last_metrics and last_metrics[-1] is None: del last_metrics[-1]
        return add_metrics(last_metrics, [float(self.epsilon)])

class RewardMetric(LearnerCallback):
    _order = -20

    def __init__(self, learn):
        super().__init__(learn)
        self.train_reward, self.valid_reward = [], []

    def on_epoch_begin(self, **kwargs:Any):
        self.train_reward, self.valid_reward = [], []

    def on_batch_end(self, **kwargs: Any):
        if self.learn.model.training: self.train_reward.append(self.learn.data.train_ds.item.reward.cpu().numpy()[0][0])
        elif not self.learn.recorder.no_val: self.valid_reward.append(self.learn.data.valid_ds.item.reward.cpu().numpy()[0][0])

    def on_train_begin(self, **kwargs):
        metric_names = ['train_reward'] if self.learn.recorder.no_val else ['train_reward', 'valid_reward']
        self.learn.recorder.add_metric_names(metric_names)

    def on_epoch_end(self, last_metrics, **kwargs: Any):
        return add_metrics(last_metrics, [sum(self.train_reward), sum(self.valid_reward)])




from fast_rl.core.agent_core import ExperienceReplay
from fast_rl.core.basic_train import AgentLearner, PipeLine
from fast_rl.core.data_block import MDPDataBunch
from fast_rl.core.train import GroupAgentInterpretation, AgentInterpretation


# def pipeline_fn(_):
#     group_interp = GroupAgentInterpretation()
#     data = MDPDataBunch.from_env('CartPole-v1', max_steps=40, render='rgb_array', bs=5, device='cpu')
#     model = DQN(data, tree=ExperienceReplay(memory_size=100, reduce_ram=True))
#     learn = AgentLearner(data, model)
#     learn.fit(2)
#     interp = AgentInterpretation(learn)
#     interp.plot_rewards(cumulative=True, per_episode=True, group_name='run', no_show=True)
#     group_interp.add_interpretation(interp)
#     data.close()
#     return group_interp.analysis
#
#
# def test_pipeline_init():
#     pl = PipeLine(2, pipeline_fn)
#     print(pl.start(5))


import numpy as np
import torch
from fastai.basic_train import LearnerCallback, Any, ifnone, listify, List
from fastai.tabular.data import emb_sz_rule

from fast_rl.agents.ddpg_models import DDPGModule
from fast_rl.core.agent_core import ExperienceReplay, ExplorationStrategy, Experience
from fast_rl.core.basic_train import AgentLearner
from fast_rl.core.data_block import MDPDataBunch, MDPStep, FEED_TYPE_STATE, FEED_TYPE_IMAGE


class DDPGLearner(AgentLearner):
    def __init__(self, data: MDPDataBunch, model, memory, exploration_method, trainers,
                 **kwargs):
        self.memory: Experience = memory
        self.exploration_method: ExplorationStrategy = exploration_method
        super().__init__(data=data, model=model, **kwargs)
        self.ddpg_trainers = listify(trainers)
        for t in self.ddpg_trainers: self.callbacks.append(t(self))

    def predict(self, element, **kwargs):
        training = self.model.training
        if element.shape[0] == 1: self.model.eval()
        pred = self.model(element)
        if training: self.model.train()
        return self.exploration_method.perturb(pred.detach().cpu().numpy(), self.data.action.action_space)

    def interpret_q(self, item):
        with torch.no_grad():
            return self.model.interpret_q(item).cpu().numpy().item()



class BaseDDPGTrainer(LearnerCallback):
    def __init__(self, learn):
        super().__init__(learn)
        self.max_episodes = 0
        self.episode = 0
        self.iteration = 0
        self.copy_over_frequency = 3

    @property
    def learn(self) -> DDPGLearner:
        return self._learn()

    def on_train_begin(self, n_epochs, **kwargs: Any):
        self.max_episodes = n_epochs

    def on_epoch_begin(self, epoch, **kwargs: Any):
        self.episode = epoch
        self.iteration = 0

    def on_loss_begin(self, **kwargs: Any):
        """Performs tree updates, exploration updates, and model optimization."""
        if self.learn.model.training: self.learn.memory.update(item=self.learn.data.x.items[-1], device=self.learn.data.device)
        self.learn.exploration_method.update(self.episode, max_episodes=self.max_episodes, explore=self.learn.model.training)
        if not self.learn.warming_up:
            samples: List[MDPStep] = self.memory.sample(self.learn.data.bs)
            post_optimize = self.learn.model.optimize(samples)
            if self.learn.model.training:
                self.learn.memory.refresh(post_optimize=post_optimize)
                self.learn.model.target_copy_over()
            self.iteration += 1


def create_ddpg_model(data: MDPDataBunch, base_arch: DDPGModule, layers=None, ignore_embed=False, channels=None,
                     opt=torch.optim.RMSprop, loss_func=None, **kwargs):
    bs, state, action = data.bs, data.state, data.action
    nc, w, h, n_conv_blocks = -1, -1, -1, [] if state.mode == FEED_TYPE_STATE else ifnone(channels, [3, 3, 1])
    if state.mode == FEED_TYPE_IMAGE: nc, w, h = state.s.shape[3], state.s.shape[2], state.s.shape[1]
    _layers = ifnone(layers, [400, 200] if len(n_conv_blocks) == 0 else [200, 200])
    if ignore_embed or np.any(state.n_possible_values == np.inf) or state.mode == FEED_TYPE_IMAGE: emb_szs = []
    else: emb_szs = [(d+1, int(emb_sz_rule(d))) for d in state.n_possible_values.reshape(-1, )]
    ao = int(action.taken_action.shape[1])
    model = base_arch(ni=state.s.shape[1], ao=ao, layers=_layers, emb_szs=emb_szs, n_conv_blocks=n_conv_blocks,
                      nc=nc, w=w, h=h, opt=opt, loss_func=loss_func, **kwargs)
    return model


ddpg_config = {
    DDPGModule: BaseDDPGTrainer
}


def ddpg_learner(data: MDPDataBunch, model, memory: ExperienceReplay, exploration_method: ExplorationStrategy,
                trainers=None, **kwargs):
    trainers = ifnone(trainers, ddpg_config[model.__class__])
    return DDPGLearner(data, model, memory, exploration_method, trainers, **kwargs)



import io
from functools import partial
from typing import List, Tuple, Dict

import matplotlib.pyplot as plt
import numpy as np
import scipy.stats as st
import torch
from PIL import Image
from fastai.train import Interpretation, DatasetType, copy
from gym.spaces import Box
from itertools import product
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from moviepy.video.VideoClip import VideoClip
from moviepy.video.io.bindings import mplfig_to_npimage
from torch import nn

from fast_rl.core import Learner
from fast_rl.core.data_block import MarkovDecisionProcessSliceAlpha, FEED_TYPE_IMAGE


class AgentInterpretationAlpha(Interpretation):
    def __init__(self, learn: Learner, ds_type: DatasetType = DatasetType.Valid, base_chart_size=(20, 10)):
        """
        Handles converting a learner, and it's runs into useful human interpretable information.

        Notes:
            This class is called AgentInterpretationAlpha because it will overall get deprecated.
            The final working version will be called AgentInterpretation.

        Args:
            learn:
        """
        super().__init__(learn, None, None, None, ds_type=ds_type)
        self.current_animation = None
        plt.rcParams["figure.figsize"] = base_chart_size
        
    def _get_items(self, ignore=True):
        episodes = list(self.ds.x.info.keys())
        if ignore or len(episodes) == 0: return self.ds.x.items
        return [item for item in self.ds.x.items if item.episode in episodes]

    @classmethod
    def from_learner(cls, learn: Learner, ds_type: DatasetType = DatasetType.Valid, activ: nn.Module = None):
        raise NotImplementedError

    def normalize(self, item: np.array):
        if np.max(item) - np.min(item) != 0:
            return np.divide(item + np.min(item), np.max(item) - np.min(item))
        else:
            item.fill(1)
            return item

    def top_losses(self, k: int = None, largest=True):
        raise NotImplementedError

    def reward_heatmap(self, episode_slices: List[MarkovDecisionProcessSliceAlpha], action=None):
        """
        Takes a state_space and uses the agent to heat map rewards over the space.

        We first need to determine if the s space is discrete or discrete.

        Args:
            state_space:

        Returns:

        """
        if action is not None: action = torch.tensor(action).long()
        current_state_slice = [p for p in product(
            np.arange(min(self.ds.env.observation_space.low), max(self.ds.env.observation_space.high) + 1),
            repeat=len(self.ds.env.observation_space.high))]
        heat_map = np.zeros(np.add(self.ds.env.observation_space.high, 1))
        with torch.no_grad():
            for state in current_state_slice:
                if action is not None:
                    heat_map[state] = self.learn.model(torch.from_numpy(np.array(state)).unsqueeze(0))[0].gather(0, action)
                else:
                    self.learn.model.eval()
                    if self.learn.model.name == 'DDPG':
                        heat_map[state] = self.learn.model.critic_model(torch.cat((torch.from_numpy(np.array(state)).unsqueeze(0).float(), self.learn.model.action_model(torch.from_numpy(np.array(state)).unsqueeze(0).float())), 1))
                    else:
                        heat_map[state] = self.learn.model(torch.from_numpy(np.array(state)).unsqueeze(0))[0].max().numpy()
        return heat_map

    def plot_heatmapped_episode(self, episode, fig_size=(13, 5), action_index=None, return_heat_maps=False):
        """
        Generates plots of heatmapped s spaces for analyzing reward distribution.

        Currently only makes sense for grid based envs. Will be expecting gym_maze environments that are discrete.

        Returns:

        """
        if not str(self.ds.env.spec).__contains__('maze'):
            raise NotImplementedError('Currently only supports gym_maze envs that have discrete s spaces')
        if not isinstance(self.ds.state_size, Box):
            raise NotImplementedError('Currently only supports Box based s spaces with 2 dimensions')

        items = self._get_items()
        heat_maps = []

        # For each episode
        buffer = []
        episode = episode if episode != -1 else list(set([i.episode for i in items]))[-1]
        for item in [i for i in items if i.episode == episode]:
            buffer.append(item)
        heat_map = self.reward_heatmap(buffer, action=action_index)
        heat_maps.append((copy(heat_map), copy(buffer[-1]), copy(episode)))

        plots = []
        for single_heatmap in [heat_maps[-1]]:
            fig, ax = plt.subplots(1, 2, figsize=fig_size)
            fig.suptitle(f'Episode {episode}')
            ax[0].imshow(single_heatmap[1].to_one().data)
            im = ax[1].imshow(single_heatmap[0])
            ax[0].grid(False)
            ax[1].grid(False)
            ax[0].set_title('Final State Snapshot')
            ax[1].set_title('State Space Heatmap')
            fig.colorbar(im, ax=ax[1])

            buf = io.BytesIO()
            fig.savefig(buf, format='png')
            # Closing the figure prevents it from being displayed directly inside
            # the notebook.
            plt.close(fig)
            buf.seek(0)
            # Create Image object
            plots.append(np.array(Image.open(buf))[:, :, :3])

        for plot in plots:
            plt.grid(False)
            plt.xticks([])
            plt.yticks([])
            plt.tight_layout()
            plt.imshow(plot)
            plt.show()

        if return_heat_maps: return heat_maps

    def plot_episode(self, episode):
        items = self._get_items(False)  # type: List[MarkovDecisionProcessSliceAlpha]

        episode_counter = 0
        # For each episode
        buffer = []
        for item in items:
            buffer.append(item)
            if item.done:
                if episode_counter == episode:
                    break
                episode_counter += 1
                buffer = []

        plots = []
        with torch.no_grad():
            agent_reward_plots = [self.learn.model(torch.from_numpy(np.array(i.current_state))).max().numpy() for i in
                                  buffer]
            fig, ax = plt.subplots(1, 1, figsize=(5, 5))
            fig.suptitle(f'Episode {episode}')
            ax.plot(agent_reward_plots)
            ax.set_xlabel('Time Steps')
            ax.set_ylabel('Max Expected Reward from Agent')

            buf = io.BytesIO()
            fig.savefig(buf, format='png')
            # Closing the figure prevents it from being displayed directly inside
            # the notebook.
            plt.close(fig)
            buf.seek(0)
            # Create Image object
            plots.append(np.array(Image.open(buf))[:, :, :3])

        for plot in plots:
            plt.grid(False)
            plt.xticks([])
            plt.yticks([])
            plt.tight_layout()
            plt.imshow(plot)
            plt.show()

    def get_agent_accuracy_density(self, items, episode_num=None):
        x = None
        y = None

        for episode in [_ for _ in list(set(mdp.episode for mdp in items)) if episode_num is None or episode_num == _]:
            subset = [item for item in items if item.episode == episode]
            state = np.array([_.current_state for _ in subset])
            result_state = np.array([_.result_state for _ in subset])

            prim_q_pred = self.learn.model(torch.from_numpy(state))
            target_q_pred = self.learn.model.target_net(torch.from_numpy(state).float())
            state_difference = (prim_q_pred - target_q_pred).sum(1)
            prim_q_pred = self.learn.model(torch.from_numpy(result_state))
            target_q_pred = self.learn.model.target_net(torch.from_numpy(result_state).float())
            result_state_difference = (prim_q_pred - target_q_pred).sum(1)

            x = state_difference if x is None else np.hstack((x, state_difference))
            y = result_state_difference if y is None else np.hstack((y, result_state_difference))

        return x, y

    def plot_agent_accuracy_density(self, episode_num=None):
        """
        Heat maps the density of actual vs estimated q v. Good reference for this is at [1].

        References:
            [1] "Simple Example Of 2D Density Plots In Python." Medium. N. p., 2019. Web. 31 Aug. 2019.
            https://towardsdatascience.com/simple-example-of-2d-density-plots-in-python-83b83b934f67

        Returns:

        """
        items = self._get_items(False)  # type: List[MarkovDecisionProcessSliceAlpha]
        x, y = self.get_agent_accuracy_density(items, episode_num)

        fig = plt.figure(figsize=(8, 8))
        ax = fig.gca()
        fig.suptitle(f'{self.learn.model.name} for {self.ds.env.spec._env_name}')
        ax.set_ylabel('State / State Prime Q Value Deviation')
        ax.set_xlabel('Iterations')
        ax.plot(np.hstack([x, y]))
        plt.show()

    def get_q_density(self, items, episode_num=None):
        x = None
        y = None

        for episode in [_ for _ in list(set(mdp.episode for mdp in items)) if episode_num is None or episode_num == _]:
            subset = [item for item in items if item.episode == episode]
            r = np.array([_.reward for _ in subset])
            # Gets the total accumulated r over a single markov chain
            actual_returns = np.flip([np.cumsum(r)[i:][0] for i in np.flip(np.arange(len(r)))]).reshape(1, -1)
            estimated_returns = self.learn.model.interpret_q(subset).view(1, -1).numpy()
            x = actual_returns if x is None else np.hstack((x, actual_returns))
            y = estimated_returns if y is None else np.hstack((y, estimated_returns))

        return self.normalize(x), self.normalize(y)

    def plot_q_density(self, episode_num=None):
        """
        Heat maps the density of actual vs estimated q v. Good reference for this is at [1].

        References:
            [1] "Simple Example Of 2D Density Plots In Python." Medium. N. p., 2019. Web. 31 Aug. 2019.
            https://towardsdatascience.com/simple-example-of-2d-density-plots-in-python-83b83b934f67

        Returns:

        """
        items = self._get_items(False)  # type: List[MarkovDecisionProcessSliceAlpha]
        x, y = self.get_q_density(items, episode_num)

        # Define the borders
        deltaX = (np.max(x) - np.min(x)) / 10
        deltaY = (np.max(y) - np.min(y)) / 10
        xmin = np.min(x) - deltaX
        xmax = np.max(x) + deltaX
        ymin = np.min(y) - deltaY
        ymax = np.max(y) + deltaY
        print(xmin, xmax, ymin, ymax)
        # Create meshgrid
        xx, yy = np.mgrid[xmin:xmax:100j, ymin:ymax:100j]

        positions = np.vstack([xx.ravel(), yy.ravel()])
        values = np.vstack([x, y])

        kernel = st.gaussian_kde(values)

        f = np.reshape(kernel(positions).T, xx.shape)

        fig = plt.figure(figsize=(8, 8))
        ax = fig.gca()
        ax.set_xlim(xmin, xmax)
        ax.set_ylim(ymin, ymax)
        cfset = ax.contourf(xx, yy, f, cmap='coolwarm')
        ax.imshow(np.rot90(f), cmap='coolwarm', extent=[xmin, xmax, ymin, ymax])
        cset = ax.contour(xx, yy, f, colors='k')
        ax.clabel(cset, inline=1, fontsize=10)
        ax.set_xlabel('Actual Returns')
        ax.set_ylabel('Estimated Q')
        if episode_num is None:
            plt.title('2D Gaussian Kernel Q Density Estimation')
        else:
            plt.title(f'2D Gaussian Kernel Q Density Estimation for episode {episode_num}')
        plt.show()

    def plot_rewards_over_iterations(self, cumulative=False, return_rewards=False):
        items = self._get_items()
        r_iter = [el.reward[0] if np.ndim(el.reward) == 0 else np.average(el.reward) for el in items]
        if cumulative: r_iter = np.cumsum(r_iter)
        fig = plt.figure(figsize=(8, 8))
        ax = fig.gca()
        fig.suptitle(f'{self.learn.model.name} for {self.ds.env.spec._env_name}')
        ax.set_ylabel('Rewards' if not cumulative else 'Cumulative Rewards')
        ax.set_xlabel('Iterations')
        ax.plot(r_iter)
        plt.show()
        if return_rewards: return r_iter

    def plot_rewards_over_episodes(self, cumulative=False, fig_size=(8, 8)):
        items = self._get_items()
        r_iter = [(el.reward[0] if np.ndim(el.reward) == 0 else np.average(el.reward), el.episode) for el in items]
        rewards, episodes = zip(*r_iter)
        if cumulative: rewards = np.cumsum(rewards)
        fig = plt.figure(figsize=(8, 8))
        ax = fig.gca()
        fig.suptitle(f'{self.learn.model.name} for {self.ds.env.spec._env_name}')
        ax.set_ylabel('Rewards' if not cumulative else 'Cumulative Rewards')
        ax.set_xlabel('Episodes')
        ax.xaxis.set_ticks([i for i, el in enumerate(episodes) if episodes[i - 1] != el or i == 0])
        ax.xaxis.set_ticklabels([el for i, el in enumerate(episodes) if episodes[i - 1] != el or i == 0])
        ax.plot(rewards)
        plt.show()

    def episode_video_frames(self, episode=None) -> Dict[str, np.array]:
        """ Returns numpy arrays representing purely episode frames. """
        items = self._get_items(False)
        if episode is None: episode_frames = {key: None for key in list(set([_.episode for _ in items]))}
        else: episode_frames = {episode: None}

        for key in episode_frames:
            if self.ds.feed_type == FEED_TYPE_IMAGE:
                episode_frames[key] = np.array([_.current_state for _ in items if key == _.episode])
            else:
                episode_frames[key] = np.array([_.alternate_state for _ in items if key == _.episode])

        return episode_frames

    def episode_to_gif(self, episode=None, path='', fps=30):
        frames = self.episode_video_frames(episode)

        for ep in frames:
            fig, ax = plt.subplots()
            animation = VideoClip(partial(self._make_frame, frames=frames[ep], axes=ax, fig=fig, title=f'Episode {ep}'),
                                  duration=frames[ep].shape[0])
            animation.write_gif(path + f'episode_{ep}.gif', fps=fps)

    def _make_frame(self, t, frames, axes, fig, title):
        axes.clear()
        fig.suptitle(title)
        axes.imshow(frames[int(t)])
        return mplfig_to_npimage(fig)

    def iplot_episode(self, episode, fps=30):
        if episode is None: raise ValueError('The episode cannot be None for jupyter display')
        x = self.episode_video_frames(episode)[episode]
        fig, ax = plt.subplots()

        self.current_animation = VideoClip(partial(self._make_frame, frames=x, axes=ax, fig=fig,
                                                   title=f'Episode {episode}'), duration=x.shape[0])
        self.current_animation.ipython_display(fps=fps, loop=True, autoplay=True)

    def get_memory_samples(self, batch_size=None, key='reward'):
        samples = self.learn.model.memory.sample(self.learn.model.batch_size if batch_size is None else batch_size)
        if not samples: raise IndexError('Your tree seems empty.')
        if batch_size is not None and batch_size > len(self.learn.model.memory):
            raise IndexError(f'Your batch size {batch_size} > the tree\'s batch size {len(self.learn.model.memory)}')
        if key not in samples[0].obj.keys(): raise ValueError(f'Key {key} not in {samples[0].obj.keys()}')
        return [s.obj[key] for s in samples]

    def plot_memory_samples(self, batch_size=None, key='reward', fig_size=(8, 8)):
        values_of_interest = self.get_memory_samples(batch_size, key)
        fig = plt.figure(figsize=fig_size)
        ax = fig.gca()
        fig.suptitle(f'{self.learn.model.name} for {self.ds.env.spec._env_name}')
        ax.set_ylabel(key)
        ax.set_xlabel('Values')
        ax.plot(values_of_interest)
        plt.show()


"""


Notices:
  [1] SumTree implementation belongs to: https://github.com/rlcode/per
  As of 8/23/2019, does not have a license provided. As another note, this code is modified.


"""

import numpy as np


class SumTree(object):
    write = 0

    def __init__(self, capacity):
        """
        Used for PER.

        References:
              [1] SumTree implementation belongs to: https://github.com/rlcode/per

        Notes:
            As of 8/23/2019, does not have a license provided. As another note, this code is modified.


        Args:
            capacity:
        """

        self.capacity = capacity
        self.tree = np.zeros(2 * capacity - 1)
        self.data = np.zeros(capacity, dtype=object)
        self.n_entries = 0

    def _propagate(self, idx, change):
        """ Update to the root node """
        parent = (idx - 1) // 2

        self.tree[parent] += change

        if (np.isscalar(parent) and parent != 0) or (not np.isscalar(parent) and all(parent != 0)):
            if not np.isscalar(parent): change[parent == 0] = 0
            self._propagate(parent, change)

    def get_left(self, index):
        return 2 * index + 1

    def get_right(self, index):
        return self.get_left(index) + 1

    def _retrieve(self, idx, s):
        """ Finds sample on leaf node """
        left = self.get_left(idx)
        right = self.get_right(idx)

        if left >= len(self.tree):
            return idx

        if s <= self.tree[left]:
            return self._retrieve(left, s)
        else:
            return self._retrieve(right, s - self.tree[left])

    def total(self):
        return self.tree[0]

    def add(self, p, data):
        """ Store priority and sample """
        idx = self.write + self.capacity - 1

        self.data[self.write] = data
        self.update(idx, p)

        self.write += 1
        if self.write >= self.capacity:
            self.write = 0

        if self.n_entries < self.capacity:
            self.n_entries += 1

    def update(self, idx, p):
        """ Update priority """
        p = p.flatten() if not np.isscalar(p) else p
        change = p - self.tree[idx]

        self.tree[idx] = p
        self._propagate(idx, change)

    def get(self, s):
        """ Get priority and sample """
        idx = self._retrieve(0, s)
        data_index = idx - self.capacity + 1

        return idx, self.tree[idx], self.data[data_index]

    def anneal_weights(self, priorities, beta):
        sampling_probabilities = priorities / self.total()
        is_weight = np.power(self.n_entries * sampling_probabilities, -beta)
        is_weight /= is_weight.max()
        return is_weight.astype(float)

    def batch_get(self, ss):
        return np.array(list(zip(*list([self.get(s) for s in ss if self.get(s)[2] != 0]))))


def print_tree(tree: SumTree):
    print('\n')
    if tree.n_entries == 0:
        print('empty')
        return

    max_d = int(np.log2(len(tree.tree)))
    string_len_max = len(str(tree.tree[-1]))

    tree_strings = []
    display_values = None
    display_indexes = None
    for layer in range(max_d + 1):
        # Get the indexes in the current layer d
        if display_indexes is None:
            display_indexes = [[0]]
        else:
            local_list = []
            for i in [_ for _ in display_indexes[-1] if _ < len(tree.tree)]:
                if tree.get_left(i) < len(tree.tree): local_list.append(tree.get_left(i))
                if tree.get_right(i) < len(tree.tree): local_list.append(tree.get_right(i))
            display_indexes.append(local_list)

    for layer in display_indexes:
        # Get the v contained in current layer d
        if display_values is None:
            display_values = [[tree.tree[i] for i in layer]]
        else:
            display_values.append([tree.tree[i] for i in layer])

    tab_sizes = []
    spacings = []
    for i, layer in enumerate(display_values):
        # for now ignore string length
        tab_sizes.append(0 if i == 0 else (tab_sizes[-1] + 1) * 2)
        spacings.append(3 if i == 0 else (spacings[-1] * 2 + 1))

    for i, layer in enumerate(display_values):
        # tree_strings.append('*' * list(reversed(tab_sizes))[i])
        values = ''.join(str(v) + ' ' * (string_len_max * list(reversed(spacings))[i]) for v in layer)
        tree_strings.append(' ' * (string_len_max * list(reversed(tab_sizes))[i]) + values)

    for tree_string in tree_strings:
        print(tree_string)



from math import floor

from fastai.basic_train import LearnerCallback
from fastai.torch_core import *
from gym.spaces import Discrete, Box

from fast_rl.core.agent_core import ExplorationStrategy
from fast_rl.core.data_block import MDPDataBunch


# class BaseAgent(nn.Module):
#     r"""
#     One of the basic differences between this model type and typical Openai models is that this will have its
#     own learner_callbacks. This is due to the often strange and beautiful methods created for training RL agents.
#     """
#     def __init__(self, data: MDPDataBunch):
#         super().__init__()
#         self.data = data
#         self.name = ''
#         # Some definition of loss needs to be implemented
#         self.loss = None
#         self.out = None
#         self.opt = None
#         self.warming_up = False
#         self.learner_callbacks = []  # type: Collection[LearnerCallback]
#         # Root model that will be accessed for action decisions
#         self.action_model = None  # type: nn.Module
#         self.exploration_strategy = ExplorationStrategy(self.training)
#
#     def forward(self, x):
#         if isinstance(x, torch.Tensor): return x.float()
#         return x
#
#     def pick_action(self, x):
#         x = self(x)
#         self.out = x
#
#         with torch.no_grad():
#             if len(x.shape) > 2: raise ValueError('The agent is outputting actions with more than 1 dimension...')
#
#             if isinstance(self.data.train_ds.env.action_space, Discrete): action = x.argmax().cpu().numpy().item()
#             elif isinstance(self.data.train_ds.env.action_space, Box) and len(x.shape) != 1: action = x.squeeze(0).cpu().numpy()
#
#             action = self.exploration_strategy.perturb(action, self.data.train_ds.env.action_space)
#
#             return action
#
#     def interpret_q(self, items):
#         raise NotImplementedError


def get_embedded(input_size, output_size, n_embeddings, n_extra_dims):
    embed = nn.Embedding(n_embeddings, n_extra_dims)
    out_size = embed(torch.clamp(torch.randn(input_size).long(), 0, 1)).shape
    return embed, np.prod(out_size)


class ToLong(nn.Module):
    def forward(self, x):
        return x.long()


class SwapImageChannel(nn.Module):
    def forward(self, x):
        return x.transpose(1, 3)


class Flatten(nn.Module):
    def forward(self, x):
        return x.view(x.size(0), -1)


def get_next_conv_shape(c_w, c_h, stride, kernel_size):
    h = floor((c_h - kernel_size - 2) / stride) + 1 # 3 convolutional layers given (3c, 640w, 640h)
    w = floor((c_w - kernel_size - 2) / stride) + 1
    return h, w


def get_conv(input_tuple, act, kernel_size, stride, n_conv_layers, layers):
    r"""
    Useful guideline for convolutional net shape change:

    Shape:
    - Input: :math:`(N, C_{in}, H_{in}, W_{in})`
    - Output: :math:`(N, C_{out}, H_{out}, W_{out})` where

      .. math::
          H_{out} = \left\lfloor\frac{H_{in}  + 2 \times \text{padding}[0] - \text{dilation}[0]
                    \times (\text{kernel\_size}[0] - 1) - 1}{\text{stride}[0]} + 1\right\rfloor

      .. math::
          W_{out} = \left\lfloor\frac{W_{in}  + 2 \times \text{padding}[1] - \text{dilation}[1]
                    \times (\text{kernel\_size}[1] - 1) - 1}{\text{stride}[1]} + 1\right\rfloor


    Args:
        input_tuple:
        act:
        kernel_size:
        stride:
        n_conv_layers:
        layers:
    """
    h, w = input_tuple[1], input_tuple[2]
    conv_layers = [SwapImageChannel()]
    for i in range(n_conv_layers):
        h, w = get_next_conv_shape(h, w, stride, kernel_size)
        conv_layers.append(torch.nn.Conv2d(input_tuple[3], 3, kernel_size=kernel_size, stride=stride))
        conv_layers.append(act)

    output_size = torch.prod(torch.tensor(nn.Sequential(*(layers + conv_layers))(torch.rand(input_tuple)).shape))
    return layers + conv_layers, output_size



from itertools import product
from time import sleep

import pytest
from fastai.basic_data import DatasetType

from fast_rl.agents.dqn import create_dqn_model, dqn_learner
from fast_rl.agents.dqn_models import *
from fast_rl.core.agent_core import ExperienceReplay, PriorityExperienceReplay, GreedyEpsilon
from fast_rl.core.data_block import MDPDataBunch, FEED_TYPE_STATE, FEED_TYPE_IMAGE
from fast_rl.core.metrics import RewardMetric, EpsilonMetric
from fast_rl.core.train import GroupAgentInterpretation, AgentInterpretation

p_model = [DQNModule, FixedTargetDQNModule, DoubleDuelingModule, DuelingDQNModule, DoubleDQNModule]
p_exp = [ExperienceReplay, PriorityExperienceReplay]
p_format = [FEED_TYPE_STATE]#, FEED_TYPE_IMAGE]
p_envs = ['CartPole-v1']

config_env_expectations = {
	'CartPole-v1': {'action_shape': (1, 2), 'state_shape': (1, 4)},
	'maze-random-5x5-v0': {'action_shape': (1, 4), 'state_shape': (1, 2)}
}


def trained_learner(model_cls, env, s_format, experience, bs, layers, memory_size=1000000, decay=0.00001,
					copy_over_frequency=300, lr: Union[float, list] = 0.001):
	memory = experience(memory_size=memory_size, reduce_ram=True)
	explore = GreedyEpsilon(epsilon_start=1, epsilon_end=0.1, decay=decay)
	if type(lr) == list: lr = lr[0] if model_cls == DQNModule else lr[2]
	model = create_dqn_model(model_cls, lr=lr, layers=layers, copy_over_frequency=copy_over_frequency)
	data = MDPDataBunch.from_env(env, render='human', bs=bs, add_valid=False, feed_type=s_format)
	learn = dqn_learner(data, model, memory=memory, exploration_method=explore,
						callback_fns=[RewardMetric, EpsilonMetric])
	learn.fit(450)
	return learn


@pytest.mark.parametrize(["model_cls", "s_format", "env"], list(product(p_model, p_format, p_envs)))
def test_dqn_create_dqn_model(model_cls, s_format, env):
	data = MDPDataBunch.from_env(env, render='rgb_array', bs=32, add_valid=False, feed_type=s_format)
	model = create_dqn_model(data, model_cls)
	model.eval()
	model(data.state.s)

	assert config_env_expectations[env]['action_shape'] == (1, data.action.n_possible_values.item())
	if s_format == FEED_TYPE_STATE:
		assert config_env_expectations[env]['state_shape'] == data.state.s.shape


@pytest.mark.parametrize(["model_cls", "s_format", "mem", "env"], list(product(p_model, p_format, p_exp, p_envs)))
def test_dqn_dqn_learner(model_cls, s_format, mem, env):
	data = MDPDataBunch.from_env(env, render='rgb_array', bs=32, add_valid=False, feed_type=s_format)
	model = create_dqn_model(data, model_cls)
	memory = mem(memory_size=1000, reduce_ram=True)
	exploration_method = GreedyEpsilon(epsilon_start=1, epsilon_end=0.1, decay=0.001)
	dqn_learner(data=data, model=model, memory=memory, exploration_method=exploration_method)

	assert config_env_expectations[env]['action_shape'] == (1, data.action.n_possible_values.item())
	if s_format == FEED_TYPE_STATE:
		assert config_env_expectations[env]['state_shape'] == data.state.s.shape


@pytest.mark.parametrize(["model_cls", "s_format", "mem", "env"], list(product(p_model, p_format, p_exp, p_envs)))
def test_dqn_fit(model_cls, s_format, mem, env):
	data = MDPDataBunch.from_env(env, render='rgb_array', bs=5, max_steps=20, add_valid=False, feed_type=s_format)
	model = create_dqn_model(data, model_cls, opt=torch.optim.RMSprop)
	memory = mem(memory_size=1000, reduce_ram=True)
	exploration_method = GreedyEpsilon(epsilon_start=1, epsilon_end=0.1, decay=0.001)
	learner = dqn_learner(data=data, model=model, memory=memory, exploration_method=exploration_method)
	learner.fit(2)

	assert config_env_expectations[env]['action_shape'] == (1, data.action.n_possible_values.item())
	if s_format == FEED_TYPE_STATE:
		assert config_env_expectations[env]['state_shape'] == data.state.s.shape


@pytest.mark.parametrize(["model_cls", "s_format", "mem"], list(product(p_model, p_format, p_exp)))
def test_dqn_fit_maze_env(model_cls, s_format, mem):
	success = False
	while not success:
		try:
			data = MDPDataBunch.from_env('maze-random-5x5-v0', render='rgb_array', bs=5, max_steps=20,
										 add_valid=False, feed_type=s_format)
			model = create_dqn_model(data, model_cls, opt=torch.optim.RMSprop)
			memory = ExperienceReplay(10000)
			exploration_method = GreedyEpsilon(epsilon_start=1, epsilon_end=0.1, decay=0.001)
			learner = dqn_learner(data=data, model=model, memory=memory, exploration_method=exploration_method)
			learner.fit(2)

			assert config_env_expectations['maze-random-5x5-v0']['action_shape'] == (
				1, data.action.n_possible_values.item())
			if s_format == FEED_TYPE_STATE:
				assert config_env_expectations['maze-random-5x5-v0']['state_shape'] == data.state.s.shape
			sleep(1)
			success = True
		except Exception as e:
			if not str(e).__contains__('Surface'):
				raise Exception


@pytest.mark.usefixtures('skip_performance_check')
@pytest.mark.parametrize(["model_cls", "s_format", 'experience'], list(product(p_model, p_format, p_exp)))
def test_dqn_models_minigrids(model_cls, s_format, experience):
	group_interp = GroupAgentInterpretation()
	for i in range(5):
		learn = trained_learner(model_cls, 'MiniGrid-FourRooms-v0', s_format, experience, bs=32, layers=[64, 64],
								memory_size=1000000, decay=0.00001)

		meta = f'{experience.__name__}_{"FEED_TYPE_STATE" if s_format == FEED_TYPE_STATE else "FEED_TYPE_IMAGE"}'
		interp = AgentInterpretation(learn, ds_type=DatasetType.Train)
		interp.plot_rewards(cumulative=True, per_episode=True, group_name=meta)
		group_interp.add_interpretation(interp)
		filename = f'{learn.model.name.lower()}_{meta}'
		group_interp.to_pickle(f'../docs_src/data/minigrid_{learn.model.name.lower()}/', filename)
		del learn


@pytest.mark.usefixtures('skip_performance_check')
@pytest.mark.parametrize(["model_cls", "s_format", 'experience'],
						 list(product(p_model, p_format, p_exp)))
def test_dqn_models_cartpole(model_cls, s_format, experience):
	group_interp = GroupAgentInterpretation()
	for i in range(5):
		learn = trained_learner(model_cls, 'CartPole-v1', s_format, experience, bs=32, layers=[64, 64],
								memory_size=1000000, decay=0.00001)

		meta = f'{experience.__name__}_{"FEED_TYPE_STATE" if s_format == FEED_TYPE_STATE else "FEED_TYPE_IMAGE"}'
		interp = AgentInterpretation(learn, ds_type=DatasetType.Train)
		interp.plot_rewards(cumulative=True, per_episode=True, group_name=meta)
		group_interp.add_interpretation(interp)
		filename = f'{learn.model.name.lower()}_{meta}'
		group_interp.to_pickle(f'../docs_src/data/cartpole_{learn.model.name.lower()}/', filename)
		del learn


@pytest.mark.usefixtures('skip_performance_check')
@pytest.mark.parametrize(["model_cls", "s_format", 'experience'], list(product(p_model, p_format, p_exp)))
def test_dqn_models_lunarlander(model_cls, s_format, experience):
	group_interp = GroupAgentInterpretation()
	for i in range(5):
		learn = trained_learner(model_cls, 'LunarLander-v2', s_format, experience, bs=32, layers=[128, 64],
								memory_size=1000000, decay=0.00001, copy_over_frequency=600, lr=[0.001, 0.00025])
		meta = f'{experience.__name__}_{"FEED_TYPE_STATE" if s_format == FEED_TYPE_STATE else "FEED_TYPE_IMAGE"}'
		interp = AgentInterpretation(learn, ds_type=DatasetType.Train)
		interp.plot_rewards(cumulative=True, per_episode=True, group_name=meta)
		group_interp.add_interpretation(interp)
		filename = f'{learn.model.name.lower()}_{meta}'
		group_interp.to_pickle(f'../docs_src/data/lunarlander_{learn.model.name.lower()}/', filename)
		del learn


@pytest.mark.usefixtures('skip_performance_check')
@pytest.mark.parametrize(["model_cls", "s_format", 'experience'], list(product(p_model, p_format, p_exp)))
def test_dqn_models_mountaincar(model_cls, s_format, experience):
	group_interp = GroupAgentInterpretation()
	for i in range(5):
		learn = trained_learner(model_cls, 'MountainCar-v0', s_format, experience, bs=32, layers=[24, 12],
								memory_size=1000000, decay=0.00001, copy_over_frequency=1000)
		meta = f'{experience.__name__}_{"FEED_TYPE_STATE" if s_format == FEED_TYPE_STATE else "FEED_TYPE_IMAGE"}'
		interp = AgentInterpretation(learn, ds_type=DatasetType.Train)
		interp.plot_rewards(cumulative=True, per_episode=True, group_name=meta)
		group_interp.add_interpretation(interp)
		filename = f'{learn.model.name.lower()}_{meta}'
		group_interp.to_pickle(f'../docs_src/data/mountaincar_{learn.model.name.lower()}/', filename)

		del learn



from fastai.basic_train import LearnerCallback
from fastai.tabular.data import emb_sz_rule

from fast_rl.agents.dqn_models import *
from fast_rl.core.agent_core import ExperienceReplay, ExplorationStrategy, Experience
from fast_rl.core.basic_train import AgentLearner
from fast_rl.core.data_block import MDPDataBunch, FEED_TYPE_STATE, FEED_TYPE_IMAGE, MDPStep


class DQNLearner(AgentLearner):
    def __init__(self, data: MDPDataBunch, model, memory, exploration_method, trainers,
                 **learn_kwargs):
        self.memory: Experience = memory
        self.exploration_method: ExplorationStrategy = exploration_method
        super().__init__(data=data, model=model, **learn_kwargs)
        self.dqn_trainers = listify(trainers)
        for t in self.dqn_trainers: self.callbacks.append(t(self))

    def predict(self, element, **kwargs):
        training = self.model.training
        if element.shape[0] == 1: self.model.eval()
        pred = self.model(element)
        if training: self.model.train()
        return self.exploration_method.perturb(torch.argmax(pred, axis=1), self.data.action.action_space)

    def interpret_q(self, item):
        with torch.no_grad():
            return torch.sum(self.model(item.s)).cpu().numpy().item()


class FixedTargetDQNTrainer(LearnerCallback):
    def __init__(self, learn, copy_over_frequency=3):
        r"""Handles updating the target model in a fixed target DQN.

        Args:
            learn: Basic Learner.
            copy_over_frequency: For every N iterations we want to update the target model.
        """
        super().__init__(learn)
        self._order = 1
        self.iteration = 0
        self.copy_over_frequency = copy_over_frequency

    def on_step_end(self, **kwargs: Any):
        self.iteration += 1
        if self.iteration % self.copy_over_frequency == 0 and self.learn.model.training:
            self.learn.model.target_copy_over()


class BaseDQNTrainer(LearnerCallback):
    def __init__(self, learn: DQNLearner, max_episodes=None):
        r"""Handles basic DQN end of step model optimization."""
        super().__init__(learn)
        self.n_skipped = 0
        self._persist = max_episodes is not None
        self.max_episodes = max_episodes
        self.episode = -1
        self.iteration = 0
        # For the callback handler
        self._order = 0
        self.previous_item = None

    @property
    def learn(self) -> DQNLearner:
        return self._learn()

    def on_train_begin(self, n_epochs, **kwargs: Any):
        self.max_episodes = n_epochs if not self._persist else self.max_episodes

    def on_epoch_begin(self, epoch, **kwargs: Any):
        self.episode = epoch if not self._persist else self.episode + 1
        self.iteration = 0

    def on_loss_begin(self, **kwargs: Any):
        r"""Performs tree updates, exploration updates, and model optimization."""
        if self.learn.model.training: self.learn.memory.update(item=self.learn.data.x.items[-1], device=self.learn.data.device)
        self.learn.exploration_method.update(self.episode, max_episodes=self.max_episodes, explore=self.learn.model.training)
        if not self.learn.warming_up:
            samples: List[MDPStep] = self.memory.sample(self.learn.data.bs)
            post_optimize = self.learn.model.optimize(samples)
            if self.learn.model.training: self.learn.memory.refresh(post_optimize=post_optimize)
            self.iteration += 1


def create_dqn_model(data: MDPDataBunch, base_arch: DQNModule, layers=None, ignore_embed=False, channels=None,
                     opt=torch.optim.RMSprop, loss_func=None, lr=0.001, **kwargs):
    bs, state, action = data.bs, data.state, data.action
    nc, w, h, n_conv_blocks = -1, -1, -1, [] if state.mode == FEED_TYPE_STATE else ifnone(channels, [3, 3, 1])
    if state.mode == FEED_TYPE_IMAGE: nc, w, h = state.s.shape[3], state.s.shape[2], state.s.shape[1]
    _layers = ifnone(layers, [64, 64])
    if ignore_embed or np.any(state.n_possible_values == np.inf) or state.mode == FEED_TYPE_IMAGE: emb_szs = []
    else: emb_szs = [(d+1, int(emb_sz_rule(d))) for d in state.n_possible_values.reshape(-1, )]
    ao = int(action.n_possible_values[0])
    model = base_arch(ni=state.s.shape[1], ao=ao, layers=_layers, emb_szs=emb_szs, n_conv_blocks=n_conv_blocks,
                      nc=nc, w=w, h=h, opt=opt, loss_func=loss_func, lr=lr, **kwargs)
    return model


dqn_config = {
    DQNModule: [BaseDQNTrainer],
    DoubleDQNModule: [BaseDQNTrainer, FixedTargetDQNTrainer],
    DuelingDQNModule: [BaseDQNTrainer],
    DoubleDuelingModule: [BaseDQNTrainer, FixedTargetDQNTrainer],
    FixedTargetDQNModule: [BaseDQNTrainer, FixedTargetDQNTrainer]
}


def dqn_learner(data: MDPDataBunch, model: DQNModule, memory: ExperienceReplay, exploration_method: ExplorationStrategy,
                trainers=None, copy_over_frequency=300, **kwargs):
    trainers = ifnone(trainers, [c if c != FixedTargetDQNTrainer else partial(c, copy_over_frequency=copy_over_frequency)
                                 for c in dqn_config[model.__class__]])
    return DQNLearner(data, model, memory, exploration_method, trainers, **kwargs)




from fast_rl.core.agent_core import ExperienceReplay
from fast_rl.core.basic_train import AgentLearner
from fast_rl.core.data_block import MDPDataBunch
from fast_rl.core.metrics import RewardMetric


# def test_metrics_reward_init():
#     data = MDPDataBunch.from_env('maze-random-5x5-v0', render='human', bs=4, max_steps=100)
#     model = FixedTargetDQN(data, tree=ExperienceReplay(1000, reduce_ram=True))
#     learn = AgentLearner(data, model, callback_fns=[RewardMetric])
#     learn.fit(3)


import os
import queue
from pathlib import Path, PosixPath


def get_absolute_path(directory, project_root='fast_rl', ignore_hidden=True, ignore_files=True):
    """
    Gets the absolute path to a directory in the project structure using depth first search.

    Args:
        directory: The name of the folder to look for.
        project_root: The project root name. Generally should not be changed.
        ignore_hidden: For future use, for now, throws error because it cannot handle hidden files.
        ignore_files: For future use, for now, throws error because it is expecting directories.
    Returns:
    """
    if not ignore_hidden:
        raise NotImplementedError('ignore_hidden is not supported currently.')
    if not ignore_files:
        raise NotImplementedError('ignore_files is not supported currently.')

    full_path = Path(__file__).parents[0]  # type: PosixPath

    # Move up the path address to the project root
    while full_path.name != project_root:
        full_path = full_path.parents[0]

    if full_path.name == directory: return str(full_path)
    # Find the path to the directory
    searched_directory = queue.Queue()
    searched_directory.put_nowait(full_path)
    # Will do a breadth first search
    while not searched_directory.empty():
        full_path_str = str(searched_directory.get_nowait())
        if os.path.exists(os.path.join(full_path_str, directory)):
            return os.path.join(full_path_str, directory)

        for inner_dir in os.listdir(full_path_str):
            if str(inner_dir).__contains__('.'):
                continue  # Directory is either a file, or hidden. Skip this

            searched_directory.put_nowait(os.path.join(full_path_str, inner_dir))

    raise IOError(f'Path to {directory} not found.')


import pickle
from copy import copy
from functools import partial
from pathlib import Path

import scipy.stats as st
from torch.distributions import Normal
from dataclasses import dataclass, field
from fastai.basic_train import *
from fastai.sixel import plot_sixel
from fastai.train import Interpretation, torch, DatasetType, defaults, ifnone, warn
import matplotlib.pyplot as plt
from fastprogress.fastprogress import IN_NOTEBOOK
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from typing import Tuple, Union, List
import pandas as pd
import numpy as np
import os
from matplotlib.ticker import MaxNLocator
from itertools import cycle, product, permutations, combinations, combinations_with_replacement, islice

from fast_rl.core.data_block import MDPList


def array_flatten(array):
    return [item for sublist in array for item in sublist]


def cumulate_squash(values: Union[list, List[list]], squash_episodes=False, cumulative=False):
    if isinstance(values[0], list):
        if squash_episodes:
            if cumulative:
                values = [np.max(np.cumsum(episode)) for episode in values]
            else:
                values = [np.max(episode) for episode in values]
        else:
            if cumulative:
                values = [np.cumsum(episode) for episode in array_flatten(values)]
            else:
                values = array_flatten(values)
    else:
        if cumulative: values = np.cumsum(values)
    return values


def group_by_episode(items: MDPList, episodes: list):
    ils = [copy(items).filter_by_func(lambda x: x.episode in episodes and x.episode == ep) for ep in episodes]
    return [il for il in ils if len(il.items) != 0]


def smooth(v, smoothing_const): return np.convolve(v, np.ones(smoothing_const), 'same') / smoothing_const


@dataclass
class GroupField:
    values: list
    model: str
    meta: str
    value_type: str
    per_episode: bool

    @property
    def analysis(self):
        return {
            'average': np.average(self.values),
            'max': np.max(self.values),
            'min': np.min(self.values),
            'type': self.value_type
        }

    @property
    def unique_tuple(self): return self.model, self.meta, self.value_type

    def __eq__(self, other):
        comp_tuple = other.unique_tuple if isinstance(other, GroupField) else other
        return all([self.unique_tuple[i] == comp_tuple[i] for i in range(len(self.unique_tuple))])

    def smooth(self, smooth_groups): self.values = smooth(self.values, smooth_groups)


class AgentInterpretation(Interpretation):
    def __init__(self, learn: Learner, ds_type: DatasetType = DatasetType.Valid, close_env=True):
        super().__init__(learn, None, None, None, ds_type=ds_type)
        self.groups = []
        if close_env: self.ds.env.close()

    def get_values(self, il: MDPList, value_name, per_episode=False):
        if per_episode:
            return [self.get_values(i, value_name) for i in group_by_episode(il, list(il.info.keys()))]
        return [i.obj[value_name] for i in il.items]

    def line_figure(self, values: Union[list, List[list]], figsize=(5, 5), cumulative=False, per_episode=False):
        fig, ax = plt.subplots(1, 1, figsize=figsize)  # type: Figure, Axes
        ax.plot(values)

        ax.set_title(f'Rewards over {"episodes" if per_episode else "iterations"}')
        ax.set_ylabel(f'{"Cumulative " if cumulative else ""}Rewards')
        ax.set_xlabel("Episodes " if per_episode else "Iterations")
        ax.xaxis.set_major_locator(MaxNLocator(integer=True))
        return fig

    def plot_rewards(self, per_episode=False, return_fig: bool = None, group_name=None, cumulative=False, no_show=False,
                     smooth_const: Union[None, float] = None, **kwargs):
        values = self.get_values(self.ds.x, 'reward', per_episode)
        processed_values = cumulate_squash(values, squash_episodes=per_episode, cumulative=cumulative, **kwargs)
        if group_name: self.groups.append(GroupField(processed_values, self.learn.model.name, group_name, 'reward',
                                                     per_episode))
        if no_show: return
        if smooth_const: processed_values = smooth(processed_values, smooth_const)
        fig = self.line_figure(processed_values, per_episode=per_episode, cumulative=cumulative, **kwargs)

        if ifnone(return_fig, defaults.return_fig): return fig
        if not IN_NOTEBOOK: plot_sixel(fig)

    def to_group_agent_interpretation(self):
        interp = GroupAgentInterpretation()
        interp.add_interpretation(self)
        return interp


@dataclass
class GroupAgentInterpretation(object):
    groups: List[GroupField] = field(default_factory=list)
    in_notebook: bool = IN_NOTEBOOK

    @property
    def analysis(self):
        if not self.in_notebook: return [g.analysis for g in self.groups]
        else: return pd.DataFrame([{'name': g.unique_tuple, **g.analysis} for g in self.groups])

    def append_meta(self, post_fix):
        r""" Useful before calling `to_pickle` if you want this set to be seen differently from future runs."""
        for g in self.groups: g.meta = g.meta + post_fix
        return self

    def filter_by(self, per_episode, value_type):
        return copy([g for g in self.groups if g.value_type == value_type and g.per_episode == per_episode])

    def group_by(self, groups, unique_values):
        for comp_tuple in unique_values: yield [g for g in groups if g == comp_tuple]

    def add_interpretation(self, interp):
        self.groups += interp.groups

    def plot_reward_bounds(self, title=None, return_fig: bool = None, per_episode=False,
                           smooth_groups: Union[None, float] = None, figsize=(5, 5)):
        groups = self.filter_by(per_episode, 'reward')
        if smooth_groups is not None: [g.smooth(smooth_groups) for g in groups]
        unique_values = list(set([g.unique_tuple for g in groups]))
        colors = list(islice(cycle(plt.rcParams['axes.prop_cycle'].by_key()['color']), len(unique_values)))
        fig, ax = plt.subplots(1, 1, figsize=figsize)  # type: Figure, Axes

        for grouped, c in zip(self.group_by(groups, unique_values), colors):
            min_len = min([len(v.values) for v in grouped])
            min_b = np.min([v.values[:min_len] for v in grouped], axis=0)
            max_b = np.max([v.values[:min_len] for v in grouped], axis=0)
            # TODO fit function sometimes does +1 more episodes...  WHY?
            overflow = [v.values for v in grouped if len(v.values) - min_len > 2]

            ax.plot(min_b, c=c, label=f'{grouped[0].meta} {grouped[0].model}')
            ax.plot(max_b, c=c)
            for v in overflow: ax.plot(v, c=c)
            ax.legend(loc='center left', bbox_to_anchor=(1, 0.5))

            ax.fill_between(list(range(min_len)), min_b, max_b, where=max_b > min_b, color=c, alpha=0.3)

        ax.xaxis.set_major_locator(MaxNLocator(integer=True))
        ax.set_title(ifnone(title, f'{"Per Episode" if per_episode else "Per Step"} Rewards'))
        ax.set_ylabel('Rewards')
        ax.set_xlabel(f'{"Episodes" if per_episode else "Steps"}')

        if ifnone(return_fig, defaults.return_fig): return fig
        if not IN_NOTEBOOK: plot_sixel(fig)

    def to_pickle(self, root_path, name):
        if not os.path.exists(root_path): os.makedirs(root_path)
        pickle.dump(self, open(Path(root_path) / (name + ".pickle"), "wb"), pickle.HIGHEST_PROTOCOL)

    @classmethod
    def from_pickle(cls, root_path, name) -> 'GroupAgentInterpretation':
        return pickle.load(open(Path(root_path) / f'{name}.pickle', 'rb'))

    def merge(self, other):
        return __class__(self.groups + other.groups)


class GymMazeInterpretation(AgentInterpretation):
    def __init__(self, learn: Learner, **kwargs):
        super().__init__(learn, **kwargs)
        try:
            from gym_maze.envs import MazeEnv
            if not issubclass(self.learn.data.env.unwrapped.__class__, MazeEnv):
                raise NotImplemented('This learner was trained on an environment that is not a gym maze env.')
        except ImportError as e:
            warn(f'Could not import gym maze. Do you have it installed? Full error: {e}')
        self.bounds = self.learn.data.state.bounds

    def eval_action(self, action_raw, index=-1):
        action_raw = action_raw[0]  # Remove batch dim
        return torch.max(action_raw).numpy().item() if index == -1 else action_raw[index].numpy().item()

    def heat_map(self, action):
        action_eval_fn = partial(self.eval_action, index=action)
        state_bounds = list(product(*(np.arange(r[0], r[1]) for r in zip(self.bounds.min, self.bounds.max))))
        # if min is -1, then it is an extra dimension, so multiply by -1 so that the dim in max + 1.
        heat_map = np.zeros(shape=tuple(self.bounds.max + -1 * self.bounds.min))
        action_map = np.zeros(shape=tuple(self.bounds.max + -1 * self.bounds.min))
        for state in state_bounds:
            with torch.no_grad():
                heat_map[state] = action_eval_fn(self.learn.model(torch.Tensor(data=state).unsqueeze(0).long()))
                action_map[state] = self.learn.predict(torch.Tensor(data=state).unsqueeze(0).long())
        return heat_map, action_map

    def add_text_to_image(self, ax, action_map):
        x_start, y_start, x_end, y_end, size = 0, 0, action_map.shape[0], action_map.shape[1], 1
        # Add the text
        jump_x = size
        jump_y = size
        x_positions = np.linspace(start=x_start, stop=x_end, num=x_end, endpoint=False) - 1
        y_positions = np.linspace(start=y_start, stop=y_end, num=y_end, endpoint=False) - 1

        for y_index, y in enumerate(y_positions):
            for x_index, x in enumerate(x_positions):
                label = action_map[y_index, x_index]
                text_x = x + jump_x
                text_y = y + jump_y
                ax.text(text_x, text_y, int(label), color='black', ha='center', va='center')

    def plot_heat_map(self, action=-1, figsize=(7, 7), return_fig=None):
        exploring = self.learn.exploration_method.explore
        self.learn.exploration_method.explore = False
        heat_map, chosen_actions = self.heat_map(action)
        fig, ax = plt.subplots(1, 1, figsize=figsize)  # type: Figure, Axes
        im = ax.imshow(heat_map)
        fig.colorbar(im)
        title = f'Heat mapped values {"maximum" if action == -1 else "for action " + str(action)}'
        title += '\nText: Chosen action for a given state'
        ax.set_title(title)
        self.add_text_to_image(ax, chosen_actions)
        self.learn.exploration_method.explore = exploring

        if ifnone(return_fig, defaults.return_fig): return fig
        if not IN_NOTEBOOK: plot_sixel(fig)


class QValueInterpretation(AgentInterpretation):
    def __init__(self, learn: Learner, **kwargs):
        super().__init__(learn, **kwargs)
        self.items = self.learn.data.x if len(self.learn.data.x) != 0 else self.learn.memory.memory

    def normalize(self, item: np.array):
        if np.max(item) - np.min(item) != 0:
            return np.divide(item + np.min(item), np.max(item) - np.min(item))
        else:
            item.fill(1)
            return item

    def q(self, items):
        actual, predicted = [], []
        episode_partition = [[i for i in items.items if i.episode == key] for key in items.info]

        for ei in episode_partition:
            if not ei: continue
            raw_actual = [ei[i].reward.cpu().numpy().item() for i in np.flip(np.arange(len(ei)))]
            actual += np.flip([np.cumsum(raw_actual[i:])[-1] for i in range(len(raw_actual))]).reshape(-1,).tolist()
            for item in ei: predicted.append(self.learn.interpret_q(item))

        return self.normalize(actual), self.normalize(predicted)

    def plot_q(self, figsize=(8, 8), return_fig=None):
        r"""
        Heat maps the density of actual vs estimated q v. Good reference for this is at [1].

        References:
            [1] "Simple Example Of 2D Density Plots In Python." Medium. N. p., 2019. Web. 31 Aug. 2019.
            https://towardsdatascience.com/simple-example-of-2d-density-plots-in-python-83b83b934f67

        Returns:

        """
        q_action, q_predicted = self.q(self.items)
        # Define the borders
        deltaX = (np.max(q_action) - np.min(q_action)) / 10
        deltaY = (np.max(q_predicted) - np.min(q_predicted)) / 10
        xmin = np.min(q_action) - deltaX
        xmax = np.max(q_action) + deltaX
        ymin = np.min(q_predicted) - deltaY
        ymax = np.max(q_predicted) + deltaY
        # Create meshgrid
        xx, yy = np.mgrid[xmin:xmax:100j, ymin:ymax:100j]

        positions = np.vstack([xx.ravel(), yy.ravel()])
        values = np.vstack([q_action, q_predicted])

        kernel = st.gaussian_kde(values)

        f = np.reshape(kernel(positions).T, xx.shape)

        fig = plt.figure(figsize=figsize)
        ax = fig.gca()
        ax.set_xlim(xmin, xmax)
        ax.set_ylim(ymin, ymax)
        cfset = ax.contourf(xx, yy, f, cmap='coolwarm')
        ax.imshow(np.rot90(f), cmap='coolwarm', extent=[xmin, xmax, ymin, ymax])
        cset = ax.contour(xx, yy, f, colors='k')
        ax.clabel(cset, inline=1, fontsize=10)
        ax.set_xlabel('Actual Returns')
        ax.set_ylabel('Estimated Q')
        ax.set_title('2D Gaussian Kernel Q Density Estimation')

        if ifnone(return_fig, defaults.return_fig): return fig
        if not IN_NOTEBOOK: plot_sixel(fig)


from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

VERSION = "0.9.6"

setup(name='fast_rl',
      version=VERSION,
      description='Fastai for computer vision and tabular learning has been amazing. One would wish that this would '
                  'be the same for RL. The purpose of this repo is to have a framework that is as easy as possible to '
                  'start, but also designed for testing new agents. ',
      url='https://github.com/josiahls/fast-reinforcement-learning',
      author='Josiah Laivins',
      author_email='jlaivins@uncc.edu',
      python_requires = '>=3.6',
      long_description=long_description,
      long_description_content_type="text/markdown",
      license='',
      packages=find_packages(),
      zip_safe=False,
      install_requires=['fastai>=1.0.59', 'gym[box2d, atari]', 'jupyter'],
      extras_require={'all': [
          'gym-minigrid',
          # 'gym_maze @ git+https://github.com/MattChanTK/gym-maze.git',
          # 'pybullet-gym @ git+https://github.com/benelot/pybullet-gym.git'
      ]},
      classifiers=[
          "Development Status :: 3 - Alpha",
          "Programming Language :: Python :: 3",
          "License :: OSI Approved :: Apache Software License",
          "Operating System :: OS Independent",
      ],
      )
